from typing import TypeAlias

_json_types: TypeAlias = dict[str, ...] | list | str | int


class _BaseExc(Exception):
    def __init__(self, msg=""):
        self.msg = msg


class ApiError(_BaseExc):
    """Raises when there is a problem with calling steam web/api methods (mostly due to `success` field),
    exclude response statuses."""

    def __init__(self, msg: str, resp: _json_types = None):
        super().__init__(msg)
        self.resp = resp


class CaptchaRequired(Exception):
    """Just when steam requires captcha, simple."""


class LoginError(ApiError):
    """When failed to do login."""


class ConfirmationError(_BaseExc):
    """Errors of all related to confirmation."""


class SessionExpired(Exception):
    """Raised when session is expired, and you need to do login"""



from typing import TYPE_CHECKING
from http.cookies import SimpleCookie
from base64 import b64encode
from time import time as time_time

from rsa import PublicKey, encrypt

from .exceptions import CaptchaRequired, LoginError, ApiError
from .constants import STEAM_URL
from .utils import get_cookie_value_from_session

if TYPE_CHECKING:
    from .client import SteamClient

__all__ = ("LoginMixin", "LOGIN_URL")

LOGIN_URL = STEAM_URL.STORE / "login"


class LoginMixin:
    """
    Mixin with confirmations related methods.
    """

    __slots__ = ()

    _is_logged: bool

    def __init__(self, *args, **kwargs):
        self._is_logged = False

        super().__init__(*args, **kwargs)

    @property
    def is_logged(self) -> bool:
        return self._is_logged

    @property
    def session_id(self: "SteamClient") -> str | None:
        return get_cookie_value_from_session(self.session, STEAM_URL.HELP.host, "sessionid")

    async def is_session_alive(self: "SteamClient") -> bool:
        r = await self.session.get(STEAM_URL.COMMUNITY)
        rt = await r.text()
        return self.username.lower() in rt.lower()

    async def login(self: "SteamClient", *, init_data=True, rsa_retries=5):
        """
        Perform login.

        :param init_data: fetch initial required data (api key, trade token)
        :param rsa_retries: retries to fetch rsa
        :raises ApiError: if failed to fetch rsa for `rsa_retries` times
        :raises LoginError:
        """

        data = await self._do_login(rsa_retries)

        # login redirects
        for url in data["transfer_urls"]:
            await self.session.post(url, data=data["transfer_parameters"])

        c_key = "sessionid"
        for url in (STEAM_URL.STORE, STEAM_URL.COMMUNITY):
            c = SimpleCookie()
            c[c_key] = self.session_id
            c[c_key]["path"] = "/"
            c[c_key]["domain"] = url.host
            c[c_key]["secure"] = True
            c[c_key]["SameSite"] = None

            self.session.cookie_jar.update_cookies(cookies=c, response_url=url)

        init_data and await self._init_data()

        self._is_logged = True

    async def _do_login(self: "SteamClient", rsa_retries: int, two_factor_code="") -> dict[str, ...]:
        pub_key, ts = await self._fetch_rsa_params(rsa_retries)
        data = {
            "password": b64encode(encrypt(self._password.encode("utf-8"), pub_key)).decode(),
            "username": self.username,
            "twofactorcode": two_factor_code,
            "emailauth": "",
            "loginfriendlyname": "",
            "captchagid": "-1",
            "captcha_text": "",
            "emailsteamid": "",
            "rsatimestamp": ts,
            "remember_login": "true",
            "donotcache": str(int(time_time() * 1000)),
        }
        r = await self.session.post(LOGIN_URL / "dologin", data=data)
        rj: dict[str, ...] = await r.json()
        if rj.get("captcha_needed"):
            raise CaptchaRequired  # need to pass args, captcha id, url and something else useful

        if rj["requires_twofactor"]:
            return await self._do_login(rsa_retries, self.two_factor_code)
        elif not rj.get("success"):
            raise LoginError(rj["message"], rj)

        return rj

    async def _fetch_rsa_params(self: "SteamClient", rsa_retries) -> tuple[PublicKey, int]:
        rj = {}
        for _ in range(rsa_retries):
            r = await self.session.post(LOGIN_URL / "getrsakey/", data={"username": self.username})
            rj = await r.json()
            try:
                rsa_mod = int(rj["publickey_mod"], 16)
                rsa_exp = int(rj["publickey_exp"], 16)
                rsa_timestamp = int(rj["timestamp"])

                return PublicKey(rsa_mod, rsa_exp), rsa_timestamp
            except KeyError:
                pass

        raise ApiError("Could not obtain rsa-key.", rj)

    async def logout(self: "SteamClient") -> None:
        await self.session.post(LOGIN_URL / "logout/", data={"sessionid": self.session_id})
        self._is_logged = False



from typing import TYPE_CHECKING
from time import time as time_time

from .utils import gen_two_factor_code, generate_confirmation_key, generate_device_id, async_throttle

if TYPE_CHECKING:
    from .client import SteamClient

__all__ = ("SteamGuardMixin",)


class SteamGuardMixin:
    """
    Mixin with Steam Guard related methods.
    """

    __slots__ = ()

    def __init__(self, *args, **kwargs):
        self._device_id = self._gen_device_id()
        super().__init__(*args, **kwargs)

    @property
    def two_factor_code(self: "SteamClient") -> str:
        return gen_two_factor_code(self._shared_secret)

    @property
    def device_id(self) -> str | None:
        return self._device_id

    @async_throttle(1, arg_name="tag")
    async def _gen_confirmation_key(self: "SteamClient", *, tag: str) -> tuple[str, int]:
        """

        :param tag:
        :return: confirmation key, used timestamp
        """
        ts = int(time_time())
        return generate_confirmation_key(self._identity_secret, tag, ts), ts

    def _gen_device_id(self: "SteamClient") -> str:
        return generate_device_id(self.steam_id)



from typing import TYPE_CHECKING, overload, Literal, Callable, TypeAlias
from json import loads
from re import compile
from datetime import datetime

from .models import Confirmation, MyMarketListing, EconItem, TradeOffer, EconItemType
from .constants import STEAM_URL, ConfirmationType, GameType, CORO
from .exceptions import ApiError, ConfirmationError, SessionExpired
from .utils import create_ident_code

if TYPE_CHECKING:
    from .client import SteamClient

__all__ = ("ConfirmationMixin", "CONF_URL")

CONF_URL = STEAM_URL.COMMUNITY / "mobileconf"
ITEM_INFO_RE = compile(r"'confiteminfo', (?P<item_info>.+), UserYou")
CONF_OP_TAGS = Literal["allow", "cancel"]
PRED: TypeAlias = Callable[[Confirmation], bool]


class ConfirmationMixin:
    """
    Mixin with confirmations related methods.
    Depends on :class:`aiosteampy.guard.SteamGuardMixin`.
    """

    __slots__ = ()

    async def remove_confirmation(self, id_or_ident: str | int, conf: Confirmation):
        """
        Remove confirmation silently from cache.

        You can override this method to provide your custom storage.
        """

    async def remove_multiple_confirmations(self, conf_ids: list[int | str], confs: list[Confirmation]):
        """
        Remove multiple confirmations silently from cache.

        You can override this method to provide your custom storage.
        """

        for index, conf_id in enumerate(conf_ids):
            await self.remove_confirmation(conf_id, confs[index])

    async def store_multiple_confirmations(self, conf_ids: list[int | str], confs: list[Confirmation]):
        """
        Cache multiple confirmations to inner store.

        You can override this method to provide your custom storage.
        """

    async def get_confirmation(self, id_or_ident: str | int) -> Confirmation | None:
        """
        Get conf from storage.

        You can override this method to provide your custom storage.
        """

    async def get_confirmations(self, predicate: PRED = None) -> list[Confirmation]:
        """
        Cached confirmations.

        You can override this method to provide your custom storage.
        """

    @overload
    async def confirm_sell_listing(self, obj: int, game: GameType) -> int:
        ...

    @overload
    async def confirm_sell_listing(self, obj: MyMarketListing | EconItemType | int) -> int:
        ...

    async def confirm_sell_listing(self, obj: MyMarketListing | EconItemType | int, game: GameType = None) -> int:
        """
        Perform sell listing confirmation.
        Pass ``game`` arg only with asset id.

        :param obj: `MyMarketListing` or `EconItem` that you listed or listing id or asset id
        :param game: `Game` or tuple with app and context id ints. Required when ``obj`` is asset id
        :return: listing id
        :raises ConfirmationError:
        """

        update = False
        if isinstance(obj, MyMarketListing):
            key = obj.id
        elif isinstance(obj, (EconItem, tuple)):
            key = create_ident_code(obj[3], obj[0], obj[1])
            update = True
        else:  # int
            if game:  # asset id & game
                key = create_ident_code(obj, *game)
                update = True
            else:  # listing id
                key = obj

        conf = await self.get_or_fetch_confirmation(key, update)
        await self.allow_confirmation(conf)

        return conf.creator_id

    async def confirm_trade_offer(self, offer: int | TradeOffer) -> int:
        """
        Perform sell trade offer confirmation.

        :param offer: trade offer id or `TradeOffer`
        :return: trade id
        :raises ConfirmationError:
        """

        conf = await self.get_or_fetch_confirmation(offer.id if isinstance(offer, TradeOffer) else offer)
        await self.allow_confirmation(conf)

        return conf.creator_id

    async def get_or_fetch_confirmation(self, key: str | int, update=False) -> Confirmation:
        conf = await self.get_confirmation(key)
        if not conf:
            confs = await self.fetch_confirmations(
                predicate=lambda c: c.creator_id == key or c.asset_ident_code == key,
                update=update,
            )
            if confs:
                conf = confs[0]
        if not conf:
            raise ConfirmationError(f"Can't find confirmation for {key} ident/trade/listing id.")

        return conf

    async def allow_all_confirmations(self, *, predicate: PRED = None) -> list[Confirmation]:
        """
        Fetch all confirmations and allow its (which passes ``predicate``) with single request to Steam.

        :param predicate: callable with single argument `Confirmation`, must return boolean
        :return: list of allowed confirmations
        :raises ConfirmationError:
        :raises ApiError:
        """

        confs = await self.fetch_confirmations(predicate=predicate)
        confs and await self.allow_multiple_confirmations(confs)
        return confs

    def allow_confirmation(self, conf: Confirmation) -> CORO[None]:
        """Shorthand for "send_confirmation(conf, 'allow')"."""

        return self.send_confirmation(conf, "allow")

    async def send_confirmation(self: "SteamClient", conf: Confirmation, tag: CONF_OP_TAGS) -> None:
        """
        Perform confirmation action. Remove passed conf from inner cache.

        :param conf: `Confirmation` that you wand to proceed
        :param tag: string literal of conf tag. Can be 'allow' or 'cancel'
        :raises ConfirmationError:
        """

        params = await self._create_confirmation_params(tag)
        params |= {"op": tag, "cid": conf.id, "ck": conf.nonce}
        r = await self.session.get(CONF_URL / "ajaxop", params=params)
        rj = await r.json()
        await self.remove_confirmation(conf.asset_ident_code or conf.creator_id, conf)  # delete before raise error

        if not rj.get("success"):
            raise ConfirmationError(
                f"Failed to perform confirmation action `{tag}` for {conf.creator_id} trade/listing id."
            )

    def allow_multiple_confirmations(self, confs: list[Confirmation]) -> CORO[None]:
        """Shorthand for "send_multiple_confirmations(conf, 'allow')"."""

        return self.send_multiple_confirmations(confs, "allow")

    async def send_multiple_confirmations(self: "SteamClient", confs: list[Confirmation], tag: CONF_OP_TAGS) -> None:
        """
        Perform confirmation action for multiple confs with single request to Steam.
        Remove passed conf from inner cache.

        :param confs: list of `Confirmation` that you wand to proceed
        :param tag: string literal of conf tag. Can be 'allow' or 'cancel'
        :raises ConfirmationError:
        """

        data = await self._create_confirmation_params(tag)
        data |= {"op": tag, "cid[]": [conf.id for conf in confs], "ck[]": [conf.nonce for conf in confs]}
        r = await self.session.post(CONF_URL / "multiajaxop", data=data)
        rj: dict[str, ...] = await r.json()
        # delete before raise error
        await self.remove_multiple_confirmations([c.asset_ident_code or c.creator_id for c in confs], confs)

        if not rj.get("success"):
            raise ConfirmationError(f"Failed to perform action `{tag}` for multiple confs.")

    async def fetch_confirmations(
        self: "SteamClient",
        *,
        predicate: PRED = None,
        update=False,
    ) -> list[Confirmation]:
        """
        Fetch confirmations, cache it and return.

        :param predicate: callable with single argument `Confirmation`, must return boolean
        :param update: fetch data and update confirmations.
            Requires to bind newly created sell listing to confirmation.
            You definitely don't need this
        :return: list of `Confirmation`
        :raises ApiError:
        """

        tag = "getlist"
        params = await self._create_confirmation_params(tag)
        r = await self.session.get(CONF_URL / tag, params=params)
        rj: dict[str, ...] = await r.json()
        if not rj.get("success"):
            raise ApiError("Can't fetch confirmations.", rj)

        confs = []
        if "conf" in rj:
            for conf_data in rj["conf"]:
                conf = Confirmation(
                    id=int(conf_data["id"]),
                    nonce=conf_data["nonce"],
                    creator_id=int(conf_data["creator_id"]),
                    creation_time=datetime.fromtimestamp(conf_data["creation_time"]),
                    type=ConfirmationType.get(conf_data["type"]),
                    icon=conf_data["icon"],
                    multi=conf_data["multi"],
                    headline=conf_data["headline"],
                    summary=conf_data["summary"][0],
                    warn=conf_data["warn"],
                )
                update and conf.type is ConfirmationType.LISTING and await self._update_confirmation(conf)
                confs.append(conf)

        await self.store_multiple_confirmations([c.creator_id for c in confs], confs)
        return [c for c in confs if predicate(c)] if predicate else confs

    async def _create_confirmation_params(self: "SteamClient", tag: str) -> dict[str, ...]:
        conf_key, ts = await self._gen_confirmation_key(tag=tag)
        return {
            "p": self.device_id,
            "a": self.steam_id,
            "k": conf_key,
            "t": ts,
            "m": "android",
            "tag": tag,
        }

    async def _update_confirmation(self: "SteamClient", conf: Confirmation):
        # create ident code from info to bind conf to listing without id (newly created)
        params = await self._create_confirmation_params(f"details{conf.id}")
        r = await self.session.get(CONF_URL / f"details/{conf.id}", params=params)
        rj = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Failed to fetch confirmation [{conf.id}] details.", rj)
        data: dict[str, ...] = loads(ITEM_INFO_RE.search(rj["html"])["item_info"])
        conf.asset_ident_code = create_ident_code(data["id"], data["appid"], data["contextid"])



"""Typed dicts for responses and methods."""

from typing import TypedDict, Literal


SUCCESS = Literal[0, 1]


class SellOrderTable(TypedDict):
    price: str
    price_with_fee: str
    quantity: str


class BuyOrderTable(TypedDict):
    price: str
    quantity: str


class ItemOrdersHistogram(TypedDict):
    success: SUCCESS
    sell_order_count: str
    sell_order_price: str
    sell_order_table: list[SellOrderTable]
    buy_order_count: str
    buy_order_price: str
    buy_order_table: list[BuyOrderTable]
    highest_buy_order: str
    lowest_sell_order: str

    # actually there is a lists, but tuple typing have fixed values
    buy_order_graph: list[tuple[float, int, str]]
    sell_order_graph: list[tuple[float, int, str]]

    graph_max_y: str
    graph_min_x: float
    graph_max_x: float
    price_prefix: str
    price_suffix: str


class Activity(TypedDict):
    type: str
    quantity: str
    price: str
    time: int
    avatar_buyer: str
    avatar_medium_buyer: str
    persona_buyer: str
    avatar_seller: str
    avatar_medium_seller: str
    persona_seller: str


class ItemOrdersActivity(TypedDict):
    success: SUCCESS
    activity: list[Activity]
    timestamp: int


class PriceOverview(TypedDict):
    success: SUCCESS
    lowest_price: str
    volume: str
    median_price: str


class TradeOffersSummary(TypedDict):
    pending_received_count: int
    new_received_count: int
    updated_received_count: int
    historical_received_count: int
    pending_sent_count: int
    newly_accepted_sent_count: int
    updated_sent_count: int
    historical_sent_count: int
    escrow_received_count: int
    escrow_sent_count: int



import asyncio
from base64 import b64decode, b64encode
from struct import pack, unpack
from time import time as time_time
from hmac import new as hmac_new
from hashlib import sha1
from functools import wraps
from typing import Callable, overload, ParamSpec, TypeVar, TYPE_CHECKING, TypeAlias
from http.cookies import SimpleCookie, Morsel

from bs4 import BeautifulSoup
from aiohttp import ClientSession
from yarl import URL

if TYPE_CHECKING:
    from .client import SteamCommunityMixin

__all__ = (
    "gen_two_factor_code",
    "generate_confirmation_key",
    "generate_device_id",
    "do_session_steam_auth",
    "get_session_cookie_jar",
    "get_cookie_value_from_session",
    "async_throttle",
    "create_ident_code",
    "account_id_to_steam_id",
    "steam_id_to_account_id",
    "to_int_boolean",
    "restore_from_cookies",
    "get_jsonable_cookies",
)


def gen_two_factor_code(shared_secret: str, timestamp: int = None) -> str:
    """Generate twofactor (onetime/TOTP) code."""

    if timestamp is None:
        timestamp = int(time_time())
    time_buffer = pack(">Q", timestamp // 30)  # pack as Big endian, uint64
    time_hmac = hmac_new(b64decode(shared_secret), time_buffer, digestmod=sha1).digest()
    begin = ord(time_hmac[19:20]) & 0xF
    full_code = unpack(">I", time_hmac[begin : begin + 4])[0] & 0x7FFFFFFF  # unpack as Big endian uint32
    chars = "23456789BCDFGHJKMNPQRTVWXY"
    code = ""

    for _ in range(5):
        full_code, i = divmod(full_code, len(chars))
        code += chars[i]

    return code


def generate_confirmation_key(identity_secret: str, tag: str, timestamp: int = None) -> str:
    if timestamp is None:
        timestamp = int(time_time())
    buff = pack(">Q", timestamp) + tag.encode("ascii")
    return b64encode(hmac_new(b64decode(identity_secret), buff, digestmod=sha1).digest()).decode()


# It works, however it's different that one generated from mobile app
def generate_device_id(steam_id: int) -> str:
    """Generate mobile android device id. Confirmation endpoints requires this."""

    hexed_steam_id = sha1(str(steam_id).encode("ascii")).hexdigest()
    return "android:" + "-".join(
        [hexed_steam_id[:8], hexed_steam_id[8:12], hexed_steam_id[12:16], hexed_steam_id[16:20], hexed_steam_id[20:32]]
    )


async def do_session_steam_auth(session: ClientSession, auth_url: str | URL):
    """
    Request auth page, find specs of steam openid and log in through steam with passed session.
    Useful when you need to log in 3rd party site trough Steam.

    :param session: just session
    :param auth_url: url to site, which redirect you to steam login page
    """

    r = await session.get(auth_url)
    rt = await r.text()

    soup = BeautifulSoup(rt, "html.parser")
    form = soup.find(id="openidForm")
    login_data = {
        "action": form.find(id="actionInput").attrs["value"],
        "openid.mode": form.find(attrs={"name": "openid.mode"}).attrs["value"],
        "openidparams": form.find(attrs={"name": "openidparams"}).attrs["value"],
        "nonce": form.find(attrs={"name": "nonce"}).attrs["value"],
    }

    await session.post("https://steamcommunity.com/openid/login", data=login_data)


def get_session_cookie_jar(session: ClientSession) -> dict[str, SimpleCookie]:
    """This function exists only to hide annoying alert "unresolved _cookies attr reference" from main base code."""

    return session.cookie_jar._cookies


def get_cookie_value_from_session(session: ClientSession, domain: str, field: str) -> str | None:
    """Just get value from session cookies."""

    jar = get_session_cookie_jar(session)
    if field in jar[domain]:
        return jar[domain][field].value


_P = ParamSpec("_P")
_R = TypeVar("_R")


@overload
def async_throttle(seconds: float, *, arg_index: int):
    ...


@overload
def async_throttle(seconds: float, *, arg_name: str):
    ...


@overload
def async_throttle(seconds: float):
    ...


def async_throttle(
    seconds: float, *, arg_index: int = None, arg_name: str = None
) -> Callable[[Callable[_P, _R]], Callable[_P, _R]]:
    """
    Prevents the decorated function from being called more than once per ``seconds``.
    Throttle (`await asyncio.sleep`) before call wrapped async func,
    when related arg was equal (same hash value) to related arg passed in previous func call
    if time between previous and current call lower than ``seconds``.
    Related arg must be hashable (can be dict key).
    Wrapped func must be async (return Coroutine).
    Throttle every func call time if related arg has been not specified.

    :param seconds: seconds during which call frequency has been limited.
    :param arg_index: index of related arg in *args tuple.
    :param arg_name: keyname of related arg in **kwargs.
    """

    # mega optimization, prevent arg checks in wrapped func call
    # I know about PEP8: E731, but this way is much shorter and readable
    if arg_index is None and arg_name is None:
        get_key = lambda _, __: None
    elif arg_index:
        get_key = lambda a, _: a[arg_index]
    else:
        get_key = lambda _, k: k[arg_name]

    ts_map: dict[..., float] = {}

    def decorator(f: Callable[_P, _R]) -> Callable[_P, _R]:
        @wraps(f)
        async def wrapper(*args: _P.args, **kwargs: _P.kwargs) -> _R:
            key = get_key(args, kwargs)
            ts = time_time()
            diff = ts - ts_map.get(key, 0)
            if diff < seconds:
                await asyncio.sleep(diff + 0.01)  # +10ms to ensure that second passed since last call
                ts_map[key] = time_time()
            else:
                ts_map[key] = ts

            return await f(*args, **kwargs)

        return wrapper

    return decorator


def create_ident_code(obj_id: int | str, app_id: int | str, context_id: int | str = None) -> str:
    """
    Create unique ident code for :class:`aiosteampy.models.EconItem` asset or item class
    (description) within whole Steam Economy.

    https://dev.doctormckay.com/topic/332-identifying-steam-items/

    :param obj_id: asset or class id of Steam Economy Item
    :param app_id: app id of Steam Game
    :param context_id: context id of Steam Game. Only for `EconItem`
    :return: ident code
    """

    code = f"{obj_id}_{app_id}"
    if context_id is not None:
        code += f"_{context_id}"

    return code


def steam_id_to_account_id(steam_id: int) -> int:
    """Convert steam id64 to steam id32."""

    return steam_id & 0xFFFFFFFF


def account_id_to_steam_id(account_id: int) -> int:
    """Convert steam id32 to steam id64."""

    return 1 << 56 | 1 << 52 | 1 << 32 | account_id


def to_int_boolean(s):
    """Convert something to 1, 0."""

    return 1 if s else 0


JSONABLE_COOKIE_JAR: TypeAlias = list[dict[str, dict[str, str, None, bool]]]


async def restore_from_cookies(
    cookies: JSONABLE_COOKIE_JAR,
    client: "SteamCommunityMixin",
    *,
    init_data=True,
    **init_kwargs,
):
    """
    Helper func. Restore client session from cookies.
    Login if session is not alive.
    """

    prepared = []
    for cookie_data in cookies:
        c = SimpleCookie()
        for k, v in cookie_data.items():
            m = Morsel()
            m._value = v.pop("value")
            m._key = v.pop("key")
            m._coded_value = v.pop("coded_value")
            m.update(v)
            c[k] = m

        prepared.append(c)

    for c in prepared:
        client.session.cookie_jar.update_cookies(c)
    if not (await client.is_session_alive()):
        await client.login(init_data=init_data, **init_kwargs)
    else:
        client._is_logged = True
        init_data and await client._init_data()


def get_jsonable_cookies(session: ClientSession) -> JSONABLE_COOKIE_JAR:
    """Extract and convert cookies to dict object."""

    cookie_jar = get_session_cookie_jar(session)
    return [
        {
            field_key: {
                "coded_value": morsel.coded_value,
                "key": morsel.key,
                "value": morsel.value,
                "expires": morsel["expires"],
                "path": morsel["path"],
                "comment": morsel["comment"],
                "domain": morsel["domain"],
                "max-age": morsel["max-age"],
                "secure": morsel["secure"],
                "httponly": morsel["httponly"],
                "version": morsel["version"],
                "samesite": morsel["samesite"],
            }
            for field_key, morsel in cookie.items()
        }
        for cookie in cookie_jar.values()
    ]



from typing import Callable, TypeAlias, overload

from aiohttp import ClientSession, ClientResponseError
from yarl import URL

from .models import (
    ItemDescription,
    ItemTag,
    ItemClass,
    EconItem,
    ItemAction,
    MarketListing,
    MarketListingItem,
)
from .constants import STEAM_URL, Game, Currency, GameType, Language
from .typed import ItemOrdersHistogram, ItemOrdersActivity, PriceOverview
from .exceptions import ApiError
from .utils import create_ident_code

INV_PAGE_SIZE = 2000  # steam new limit rule
INVENTORY_URL = STEAM_URL.COMMUNITY / "inventory"
PREDICATE: TypeAlias = Callable[[EconItem], bool]
PRIVATE_USER_EXC_MSG = "User inventory is private."
ITEM_MARKET_LISTINGS_DATA: TypeAlias = tuple[list[MarketListing], int]


class SteamPublicMixin:
    """Mixin contain methods that do not need authorization."""

    __slots__ = ()

    session: ClientSession
    language: Language
    currency: Currency
    country: str

    async def get_user_inventory(
        self,
        steam_id: int,
        game: GameType,
        *,
        predicate: PREDICATE = None,
        page_size=INV_PAGE_SIZE,
    ) -> list[EconItem]:
        """
        Fetches inventory of user.

        :param steam_id: steamid64 of user
        :param game: just Steam Game
        :param page_size: max items on page. Current Steam limit is 2000
        :param predicate: callable with single arg `EconItem`, must return bool
        :return: list of `EconItem`
        :raises ApiError: if response data `success` is False or user inventory is private
        """

        inv_url = INVENTORY_URL / f"{steam_id}/"
        params = {"l": self.language, "count": page_size}
        headers = {"Referer": str(inv_url)}
        url = inv_url / f"{game[0]}/{game[1]}"

        classes_map = {}  # shared classes within whole game context inventory
        items = []
        more_items = True
        last_assetid = None
        while more_items:
            params_pag = {**params, "start_assetid": last_assetid} if last_assetid else params
            data = await self._fetch_inventory(url, params_pag, headers)
            more_items = data.get("more_items", False)
            if more_items:
                last_assetid = data.get("last_assetid")

            items.extend(self._parse_items(data, steam_id, classes_map))

        return [i for i in items if predicate(i)] if predicate else items

    async def _fetch_inventory(
        self,
        url: URL,
        params: dict,
        headers: dict,
    ) -> dict[str, list[dict] | int]:
        try:
            r = await self.session.get(url, params=params, headers=headers)
        except ClientResponseError as e:
            raise ApiError(PRIVATE_USER_EXC_MSG, str(url)) if e.status == 403 else e

        rj: dict[str, list[dict] | int] = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Can't fetch inventory.", rj)

        return rj

    @staticmethod
    def _find_game_for_asset(description_data: dict[str, int], assets: list[dict[str, int | str]]) -> GameType:
        try:
            return Game(description_data["appid"])
        except ValueError:
            for asset in assets:
                if asset["classid"] == description_data["classid"]:
                    return asset["appid"], int(asset["contextid"])

    @classmethod
    def _parse_items(
        cls,
        data: dict[str, list[dict]],
        steam_id: int,
        classes_map: dict[str, ItemClass],
    ) -> tuple[EconItem, ...]:
        for d_data in data["descriptions"]:
            key = d_data["classid"]
            if key not in classes_map:
                classes_map[key] = cls._create_item_class_from_data(d_data, data["assets"])

        return tuple(
            EconItem(
                id=int(asset_data["assetid"]),
                owner_id=steam_id,
                class_=classes_map[asset_data["classid"]],
                amount=int(asset_data["amount"]),
            )
            for asset_data in data["assets"]
        )

    @classmethod
    def _create_item_class_from_data(cls, data: dict, assets: list[dict[str, int | str]]) -> ItemClass:
        return ItemClass(
            id=int(data["classid"]),
            instance_id=int(data["instanceid"]),
            game=cls._find_game_for_asset(data, assets),
            name=data["name"],
            market_name=data["market_name"],
            market_hash_name=data["market_hash_name"],
            name_color=data["name_color"] or None,
            background_color=data.get("name_color") or None,
            type=data["type"] or None,
            icon=data["icon_url"],
            icon_large=data.get("icon_url_large"),
            commodity=bool(data["commodity"]),
            tradable=bool(data["tradable"]),
            marketable=bool(data["marketable"]),
            market_tradable_restriction=data.get("market_tradable_restriction"),
            market_buy_country_restriction=data.get("market_buy_country_restriction"),
            market_fee_app=data.get("market_fee_app"),
            market_marketable_restriction=data.get("market_marketable_restriction"),
            actions=tuple(ItemAction(a_data["link"], a_data["name"]) for a_data in data.get("actions", ())),
            market_actions=tuple(
                ItemAction(a_data["link"], a_data["name"]) for a_data in data.get("market_actions", ())
            ),
            owner_actions=tuple(ItemAction(a_data["link"], a_data["name"]) for a_data in data.get("owner_actions", ())),
            tags=tuple(
                ItemTag(
                    category=t_data["category"],
                    internal_name=t_data["internal_name"],
                    localized_category_name=t_data["localized_category_name"],
                    localized_tag_name=t_data["localized_tag_name"],
                    color=t_data.get("color"),
                )
                for t_data in data.get("tags", ())
            ),
            descriptions=tuple(
                ItemDescription(
                    value=de_data["value"],
                    color=de_data.get("color"),
                )
                for de_data in data.get("descriptions", ())
                if de_data["value"] != " "  # ha, surprise!
            ),
            fraud_warnings=tuple(data.get("fraudwarnings", ())),
        )

    async def fetch_item_orders_histogram(
        self,
        item_nameid: int,
        *,
        lang: Language = None,
        country: str = None,
        currency: Currency = None,
    ) -> ItemOrdersHistogram:
        """
        Do what described in method name.

        .. seealso::
            * https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Item-Orders-Histogram
            * https://github.com/somespecialone/steam-item-name-ids

        .. warning:: This request is rate limited by Steam.

        :param item_nameid: special id of item class. Can be found only on listings page.
        :param lang:
        :param country:
        :param currency:
        :return: `ItemOrdersHistogram` dict
        :raises ApiError:
        """

        params = {
            "norender": 1,
            "language": lang or self.language,
            "country": country or self.country,
            "currency": currency or self.currency,
            "item_nameid": item_nameid,
        }
        r = await self.session.get(STEAM_URL.MARKET / "itemordershistogram", params=params)
        rj: ItemOrdersHistogram = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Can't fetch item orders histogram for {item_nameid}.", rj)

        return rj

    async def fetch_item_orders_activity(
        self,
        item_name_id: int,
        *,
        lang: Language = None,
        country: str = None,
        currency: Currency = None,
    ) -> ItemOrdersActivity:
        """
        Do what described in method name.

        .. seealso::
            * https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Item-Orders-Activity
            * https://github.com/somespecialone/steam-item-name-ids

        :param item_name_id: special id of item class. Can be found only on listings page.
        :param lang:
        :param country:
        :param currency:
        :return: `ItemOrdersActivity` dict
        :raises ApiError:
        """

        params = {
            "norender": 1,
            "language": lang or self.language,
            "country": country or self.country,
            "currency": currency or self.currency,
            "item_nameid": item_name_id,
        }
        r = await self.session.get(STEAM_URL.MARKET / "itemordersactivity", params=params)
        rj: ItemOrdersActivity = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Can't fetch item orders activity for {item_name_id}.", rj)

        return rj

    @overload
    async def fetch_price_overview(
        self,
        obj: EconItem | ItemClass,
        *,
        country: str = ...,
        currency: Currency = ...,
    ) -> PriceOverview:
        ...

    @overload
    async def fetch_price_overview(
        self,
        obj: str,
        app_id: int,
        *,
        country: str = ...,
        currency: Currency = ...,
    ) -> PriceOverview:
        ...

    async def fetch_price_overview(
        self,
        obj: str | EconItem | ItemClass,
        app_id: int = None,
        *,
        country: str = None,
        currency: Currency = None,
    ) -> PriceOverview:
        """
        Fetch price data.

        .. warning:: This request is rate limited by Steam.

        :param obj:
        :param app_id:
        :param country:
        :param currency:
        :return: `PriceOverview` dict
        :raises ApiError:
        """

        if isinstance(obj, EconItem):
            name = obj.class_.market_hash_name
            app_id = obj.class_.game.app_id
        elif isinstance(obj, ItemClass):
            name = obj.market_hash_name
            app_id = obj.game.app_id
        else:  # str
            name = obj

        params = {
            "country": country or self.country,
            "currency": currency or self.currency,
            "market_hash_name": name,
            "appid": app_id,
        }
        r = await self.session.get(STEAM_URL.MARKET / "priceoverview", params=params)
        rj: PriceOverview = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Can't fetch price overview for `{name}`.", rj)

        return rj

    @overload
    async def get_item_listings(
        self,
        obj: EconItem | ItemClass,
        *,
        country: str = ...,
        currency: Currency = ...,
        query: str = ...,
        start: int = ...,
        count: int = ...,
    ) -> ITEM_MARKET_LISTINGS_DATA:
        ...

    @overload
    async def get_item_listings(
        self,
        obj: str,
        app_id: int,
        *,
        country: str = None,
        currency: Currency = None,
        query: str = None,
        start: int = 0,
        count: int = 100,
    ) -> ITEM_MARKET_LISTINGS_DATA:
        ...

    async def get_item_listings(
        self,
        obj: str | EconItem | ItemClass,
        app_id: int = None,
        *,
        country: str = None,
        currency: Currency = None,
        lang: str = None,
        query: str = None,
        start: int = 0,
        count: int = 100,
    ) -> ITEM_MARKET_LISTINGS_DATA:
        """
        Fetch item listings from market.
        You can paginate by yourself passing ``start`` arg.

        .. warning:: This request is rate limited by Steam.

        :param obj:
        :param app_id:
        :param country:
        :param currency:
        :param lang:
        :param count:
        :param start:
        :param query:
        :return: list of `MarketListing`, total listings count
        :raises ApiError:
        """

        if isinstance(obj, EconItem):
            name = obj.class_.market_hash_name
            app_id = obj.class_.game.app_id
        elif isinstance(obj, ItemClass):
            name = obj.market_hash_name
            app_id = obj.game.app_id
        else:  # str
            name = obj

        base_url = STEAM_URL.MARKET / f"listings/{app_id}/{name}"
        params = {
            "query": query or "",
            "country": country or self.country,
            "currency": currency or self.currency,
            "start": start,
            "count": count,
            "language": lang or self.language,
        }
        headers = {"Referer": str(base_url)}
        r = await self.session.get(base_url / "render", params=params, headers=headers)
        rj: dict[str, int | dict[str, dict[str, ...]]] = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Can't fetch market listings for `{name}`.", rj)
        if not rj["total_count"] or not rj["assets"]:
            return [], 0

        # all listings on page DOES NOT have single item class, because Steam
        classes_map = {}
        assets_map = {}
        self._update_classes_map_for_public(rj["assets"], classes_map)
        self._parse_assets_for_listings(rj["assets"], classes_map, assets_map)

        return [
            MarketListing(
                id=int(l_data["listingid"]),
                item=assets_map[
                    create_ident_code(
                        l_data["asset"]["id"],
                        l_data["asset"]["appid"],
                        l_data["asset"]["contextid"],
                    )
                ],
                currency=Currency(int(l_data["currencyid"]) - 2000),
                price=int(l_data["price"]) / 100,
                fee=int(l_data["fee"]) / 100,
                converted_currency=Currency(int(l_data["converted_currencyid"]) - 2000),
                converted_fee=int(l_data["converted_fee"]) / 100,
                converted_price=int(l_data["converted_price"]) / 100,
            )
            for l_data in rj["listinginfo"].values()
        ], rj["total_count"]

    @classmethod
    def _update_classes_map_for_public(
        cls,
        assets: dict[str, dict[str, dict[str, dict[str, ...]]]],
        classes_map: dict[str, ItemClass],
    ):
        for app_id, app_data in assets.items():
            for context_id, context_data in app_data.items():
                for asset_id, a_data in context_data.items():
                    key = create_ident_code(a_data["classid"], app_id)
                    classes_map[key] = cls._create_item_class_from_data(a_data, (a_data,))

    @staticmethod
    def _parse_assets_for_listings(
        data: dict[str, dict[str, dict[str, dict[str, ...]]]],
        classes_map: dict[str, ItemClass],
        assets_map: dict[str, MarketListingItem],
    ):
        for app_id, app_data in data.items():
            for context_id, context_data in app_data.items():
                for a_data in context_data.values():
                    key = create_ident_code(a_data["id"], app_id, context_id)
                    if key not in assets_map:
                        assets_map[key] = MarketListingItem(
                            id=int(a_data["id"]),
                            market_id=0,  # market listing post init
                            class_=classes_map[create_ident_code(a_data["classid"], app_id)],
                            unowned_id=int(a_data["unowned_id"]),
                            unowned_context_id=int(a_data["unowned_contextid"]),
                        )



from dataclasses import dataclass, field
from typing import Literal, TypeAlias
from datetime import datetime

from .constants import (
    STEAM_URL,
    GameType,
    Currency,
    ConfirmationType,
    MarketHistoryEventType,
    MarketListingStatus,
    TradeOfferStatus,
)
from .utils import create_ident_code, account_id_to_steam_id


@dataclass(eq=False, slots=True)
class ItemAction:
    link: str
    name: str


@dataclass(eq=False, slots=True)
class ItemDescription:
    value: str
    type: Literal["html"] = "html"  # just because
    color: str | None = None  # hexadecimal


@dataclass(eq=False, slots=True)
class ItemTag:
    category: str
    internal_name: str
    localized_category_name: str
    localized_tag_name: str
    color: str | None = None  # hexadecimal


@dataclass(eq=False, slots=True)
class ItemClass:
    """
    `EconItem` description representation.
    `ident_code` field is guaranteed unique within whole Steam Economy.
    """

    id: int  # classid
    instance_id: int

    game: GameType

    name: str
    market_name: str
    market_hash_name: str

    type: str | None

    name_color: str | None  # hexadecimal
    background_color: str | None

    icon: str
    icon_large: str | None

    actions: tuple[ItemAction, ...]
    market_actions: tuple[ItemAction, ...]
    owner_actions: tuple[ItemAction, ...]
    tags: tuple[ItemTag, ...]
    descriptions: tuple[ItemDescription, ...]

    fraud_warnings: tuple[str]

    commodity: bool  # ?
    tradable: bool
    marketable: bool
    market_tradable_restriction: int | None = None
    market_buy_country_restriction: str | None = None
    market_fee_app: int | None = None
    market_marketable_restriction: int | None = None

    # optimization 🚀
    ident_code: str = field(init=False, default=None)

    def __post_init__(self):
        self._set_ident_code()

    def _set_ident_code(self):
        self.ident_code = create_ident_code(self.id, self.game[0])

    @property
    def d_id(self) -> int | None:
        """Optional CSGO attr"""
        for action in self.actions:
            if "Inspect" in action.name:
                return int(action.link.split("%D")[1])

    @property
    def class_id(self) -> int:
        return self.id

    @property
    def icon_url(self) -> str:
        return str(STEAM_URL.STATIC / f"economy/image/{self.icon}/96fx96f")

    @property
    def icon_large_url(self) -> str | None:
        return str(STEAM_URL.STATIC / f"economy/image/{self.icon_large}/330x192") if self.icon_large else None

    def __eq__(self, other: "ItemClass"):
        return (self.id == other.id) and (self.game[0] == other.game[0]) and (self.game[1] == other.game[1])

    def __hash__(self):
        return self.id


EconItemTuple: TypeAlias = tuple[int, int, int, int]


@dataclass(eq=False, slots=True, kw_only=True)
class EconItem:
    """
    Represents Steam economy item (inventories).
    `ident_code` field is guaranteed unique within whole Steam Economy.
    """

    id: int  # The item's unique ID within its app+context.
    owner_id: int

    class_: ItemClass

    amount: int

    ident_code: str = field(init=False, default=None)  # optimization 🚀
    _args_tuple_: EconItemTuple = field(init=False, default=(), repr=False)

    def __post_init__(self):
        self._set_ident_code()
        self._set_args_tuple()

    def _set_ident_code(self):
        self.ident_code = create_ident_code(self.id, *self.class_.game)

    def _set_args_tuple(self):
        self._args_tuple_ = (self.class_.game[0], self.class_.game[1], self.amount, self.id)

    @property
    def inspect_link(self) -> str | None:
        for action in self.class_.actions:
            if "Inspect" in action.name:
                return action.link.replace("%assetid%", str(self.id)).replace("%owner_steamid%", str(self.owner_id))

    @property
    def asset_id(self) -> int:
        return self.id

    def __getitem__(self, index: int) -> int:  # some magic
        return self._args_tuple_[index]

    def __iter__(self):
        return iter(self._args_tuple_)

    def __eq__(self, other: "EconItem"):
        return (self.id == other.id) and (self.class_ == other.class_)

    def __hash__(self):
        return hash(self.ident_code)


EconItemType: TypeAlias = EconItem | EconItemTuple


# https://github.com/DoctorMcKay/node-steamcommunity/wiki/CConfirmation
@dataclass(eq=False, slots=True)
class Confirmation:
    id: int
    nonce: str  # conf key
    creator_id: int  # trade/listing id
    creation_time: datetime

    type: ConfirmationType

    icon: str
    multi: bool  # ?
    headline: str
    summary: str
    warn: str | None  # ?

    asset_ident_code: str | None = None  # only to map confirmation to sell listing without id


@dataclass(eq=False, slots=True)
class Notifications:
    trades: int = 0  # 1
    game_turns: int = 0  # 2
    moderator_messages: int = 0  # 3
    comments: int = 0  # 4
    items: int = 0  # 5
    invites: int = 0  # 6
    # 7 missing
    gifts: int = 0  # 8
    chats: int = 0  # 9
    help_request_replies: int = 0  # 10
    account_alerts: int = 0  # 11


@dataclass(eq=False, slots=True, kw_only=True)
class MarketListingItem(EconItem):
    # presented only on active listing
    market_id: int  # listing id

    unowned_id: int | None
    unowned_context_id: int | None

    amount: int = 1  # item on listing always have 1 amount
    owner_id: int = 0

    @property
    def inspect_link(self) -> str | None:
        for action in self.class_.market_actions:
            if "Inspect" in action.name:
                return action.link.replace("%assetid%", str(self.id)).replace("%listingid%", str(self.market_id))


@dataclass(eq=False, slots=True)
class BaseOrder:
    id: int  # listing/buy order id

    price: float

    def __hash__(self):
        return self.id


@dataclass(eq=False, slots=True)
class MyMarketListing(BaseOrder):
    lister_steam_id: int
    time_created: datetime

    item: MarketListingItem

    status: MarketListingStatus
    active: bool  # ?

    # fields that can be useful
    item_expired: int
    cancel_reason: int
    time_finish_hold: int

    @property
    def listing_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True)
class BuyOrder(BaseOrder):
    item_class: ItemClass

    quantity: int
    quantity_remaining: int

    @property
    def buy_order_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True, kw_only=True)
class MarketHistoryListingItem(MarketListingItem):
    market_id: None = None

    # purchase fields
    new_id: int | None = None
    new_context_id: int | None = None
    rollback_new_id: int | None = None
    rollback_new_context_id: int | None = None

    @property
    def inspect_link(self) -> None:
        """Always `None`, because we can't be sure that asset id has not been changed."""
        return None


@dataclass(eq=False, slots=True, kw_only=True)
class MarketHistoryListing(BaseOrder):
    item: MarketHistoryListingItem

    price: float | None = None

    # purchase fields
    purchase_id: int | None = None
    steamid_purchaser: int | None = None
    received_amount: float | None = None

    # listing fields
    original_price: float | None = None
    cancel_reason: str | None = None

    @property
    def listing_id(self) -> int:
        return self.id


@dataclass(eq=False, slots=True)
class MarketHistoryEvent:
    """
    Event entity in market history. Represents event linked with related listing.
    Note that this is just a snapshot of listing, asset data for event fire moment time
    and `asset id`, `context id` of asset may change already.
    """

    listing: MarketHistoryListing
    time_event: datetime
    type: MarketHistoryEventType


@dataclass(eq=False, slots=True)
class PriceHistoryEntry:
    date: datetime
    price: float
    daily_volume: int


@dataclass(eq=False, slots=True)
class MarketListing(BaseOrder):
    item: MarketListingItem

    currency: Currency  # original currency
    fee: float

    converted_currency: Currency
    converted_price: float
    converted_fee: float

    def __post_init__(self):
        if not self.item.market_id:
            self.item.market_id = self.id

    @property
    def listing_id(self) -> int:
        return self.id

    @property
    def total_price(self) -> float:
        return self.price + self.fee

    @property
    def total_converted_price(self) -> float:
        return self.converted_price + self.converted_fee


@dataclass(eq=False, slots=True, kw_only=True)
class BaseTradeOfferItem(EconItem):
    class_id: int
    game: GameType

    class_: ItemClass | None  # old trades don't have descriptions for asset due to Steam :)

    def _set_ident_code(self):
        self.ident_code = create_ident_code(self.id, *self.game)

    def _set_args_tuple(self):
        self._args_tuple_ = (self.game[0], self.game[1], self.amount, self.id)

    @property
    def inspect_link(self) -> None:
        return None


@dataclass(eq=False, slots=True, kw_only=True)
class TradeOfferItem(BaseTradeOfferItem):
    missing: bool
    est_usd: int

    owner_id: int = 0

    def __eq__(self, other: "EconItem | TradeOfferItem"):
        if isinstance(other, EconItem):
            game = other.class_.game
            class_id = other.class_.id
        else:
            game = other.game
            class_id = other.class_id

        return (
            (self.id == other.id)
            and (self.game[0] == game[0])
            and (self.game[1] == game[1])
            and (self.class_id == class_id)
        )


@dataclass(eq=False, slots=True)
class BaseTradeOffer:
    id: int

    owner: int  # steam id64 of entity owner acc (SteamClient)
    partner_id: int  # id32

    status: TradeOfferStatus

    @property
    def trade_offer_id(self) -> int:
        return self.id

    @property
    def partner_id64(self) -> int:
        return account_id_to_steam_id(self.partner_id)

    @property
    def is_active(self) -> bool:
        return self.status is TradeOfferStatus.ACTIVE


@dataclass(eq=False, slots=True)
class TradeOffer(BaseTradeOffer):
    """Steam Trade Offer entity."""

    is_our_offer: bool

    expiration_time: datetime
    time_created: datetime
    time_updated: datetime

    items_to_give: list[TradeOfferItem]
    items_to_receive: list[TradeOfferItem]

    message: str = ""

    def __post_init__(self):
        for i in self.items_to_give:
            i.owner_id = self.owner

        for i in self.items_to_receive:
            i.owner_id = self.partner_id64

    @property
    def sender(self) -> int:
        """Steam id64 of sender."""
        return self.owner if self.is_our_offer else self.partner_id64

    @property
    def receiver(self) -> int:
        """Steam id64 of receiver."""
        return self.partner_id64 if self.is_our_offer else self.owner


@dataclass(eq=False, slots=True, kw_only=True)
class HistoryTradeOfferItem(BaseTradeOfferItem):
    new_asset_id: int
    new_context_id: int

    owner_id: None = None


@dataclass(eq=False, slots=True)
class HistoryTradeOffer(BaseTradeOffer):
    time_init: datetime

    assets_received: tuple[HistoryTradeOfferItem, ...]
    assets_given: tuple[HistoryTradeOfferItem, ...]



from typing import TYPE_CHECKING, overload, Literal, TypeAlias, Type, Callable
from datetime import datetime
from math import floor

from aiohttp import ClientResponseError
from yarl import URL

from .exceptions import ApiError, SessionExpired
from .models import (
    EconItem,
    MyMarketListing,
    BuyOrder,
    ItemClass,
    MarketListingItem,
    MarketHistoryEvent,
    MarketHistoryListing,
    MarketHistoryListingItem,
    PriceHistoryEntry,
    MarketListing,
)
from .constants import STEAM_URL, GameType, MarketListingStatus, MarketHistoryEventType
from .utils import create_ident_code

if TYPE_CHECKING:
    from .client import SteamClient

__all__ = ("MarketMixin",)

MY_LISTINGS: TypeAlias = tuple[list[MyMarketListing], list[MyMarketListing], list[BuyOrder]]
PREDICATE: TypeAlias = Callable[[MarketHistoryEvent], bool]
PRICE_HISTORY_ENTRY_DATE_FORMAT = "%b %d %Y"


class MarketMixin:
    """
    Mixin with market related methods.

    Depends on :class:`aiosteampy.confirmation.ConfirmationMixin`,
    :class:`aiosteampy.public.SteamPublicMixin`.
    """

    __slots__ = ()

    @overload
    async def place_sell_listing(self, asset: EconItem, *, price: float) -> int:
        ...

    @overload
    async def place_sell_listing(self, asset: EconItem, *, price: float, confirm: Literal[False] = ...) -> None:
        ...

    @overload
    async def place_sell_listing(self, asset: EconItem, *, to_receive: float) -> int:
        ...

    @overload
    async def place_sell_listing(self, asset: EconItem, *, to_receive: float, confirm: Literal[False] = ...) -> None:
        ...

    @overload
    async def place_sell_listing(self, asset: int, game: GameType, *, price: float) -> int:
        ...

    @overload
    async def place_sell_listing(
        self, asset: int, game: GameType, *, price: float, confirm: Literal[False] = ...
    ) -> None:
        ...

    @overload
    async def place_sell_listing(self, asset: int, game: GameType, *, to_receive: float) -> int:
        ...

    @overload
    async def place_sell_listing(
        self, asset: int, game: GameType, *, to_receive: float, confirm: Literal[False] = ...
    ) -> None:
        ...

    async def place_sell_listing(
        self: "SteamClient",
        asset: EconItem | int,
        game: GameType = None,
        *,
        price: float = None,
        to_receive: float = None,
        confirm=True,
    ) -> int | None:
        """
        Create and place sell listing.
        If `confirm` is True - return listing id of created and confirmed sell listing,
        if this requires confirmation. If not - return None.
        Money should be only and only in account wallet currency.

        :param asset: `EconItem` that you want to list on market or asset id
        :param game: game of item
        :param price: money that buyer must pay
        :param to_receive: money that you wand to receive
        :param confirm: confirm listing or not if steam demands mobile confirmation
        :return: sell listing id or None
        :raises ApiError:
        """

        if isinstance(asset, EconItem):
            asset_id = asset.id
            game = asset.class_.game
        else:
            asset_id = asset

        if to_receive is None:
            to_receive = price * (1 - (self._steam_fee + self._publisher_fee))

        data = {
            "assetid": asset_id,
            "sessionid": self.session_id,
            "contextid": game[1],
            "appid": game[0],
            "amount": 1,
            "price": int(to_receive * 100),
        }
        headers = {"Referer": str(STEAM_URL.COMMUNITY / f"profiles/{self.steam_id}/inventory")}
        r = await self.session.post(STEAM_URL.MARKET / "sellitem/", data=data, headers=headers)
        rj: dict[str, ...] = await r.json()
        if not rj.get("success"):
            raise ApiError("Failed to place sell listing.", rj)

        if rj.get("needs_mobile_confirmation") and confirm:
            return await self.confirm_sell_listing(asset_id, game)

    def cancel_sell_listing(self: "SteamClient", obj: MyMarketListing | int):
        """
        Just cancel sell listing.

        :param obj: `MyMarketListing` or listing id
        """

        listing_id: int = obj.id if isinstance(obj, MyMarketListing) else obj
        data = {"sessionid": self.session_id}
        headers = {"Referer": str(STEAM_URL.MARKET)}
        return self.session.post(STEAM_URL.MARKET / f"removelisting/{listing_id}", data=data, headers=headers)

    @overload
    async def place_buy_order(self, obj: ItemClass, *, price: float, quantity: int) -> int:
        ...

    @overload
    async def place_buy_order(self, obj: str, game: GameType, *, price: float, quantity: int) -> int:
        ...

    async def place_buy_order(
        self: "SteamClient",
        obj: str | ItemClass,
        game: GameType = None,
        *,
        price: float,
        quantity: int,
    ) -> int:
        """
        Place buy order on market.

        :param obj: `ItemClass` or market hash name
        :param game: `Game` if `obj` is market hash name
        :param price: price of single item
        :param quantity: just quantity
        :return: buy order id
        :raises ApiError:
        """

        if isinstance(obj, ItemClass):
            name = obj.market_hash_name
            game = obj.game
        else:
            name = obj

        data = {
            "sessionid": self.session_id,
            "currency": self._wallet_currency.value,
            "appid": game[0],
            "market_hash_name": name,
            "price_total": int(price * quantity * 100),
            "quantity": quantity,
        }

        headers = {"Referer": str(STEAM_URL.MARKET / f"listings/{game[0]}/{name}")}
        r = await self.session.post(STEAM_URL.MARKET / "createbuyorder/", data=data, headers=headers)
        rj: dict[str, ...] = await r.json()
        if not rj.get("success"):
            raise ApiError("Failed to create buy order.", rj)

        return int(rj["buy_orderid"])

    async def cancel_buy_order(self: "SteamClient", order: int | BuyOrder):
        """
        Just cancel buy order.

        :param order: `BuyOrder` or buy order id
        :raises ApiError:
        """

        if isinstance(order, BuyOrder):
            order_id = order.id
        else:
            order_id = order

        data = {"sessionid": self.session_id, "buy_orderid": order_id}
        headers = {"Referer": str(STEAM_URL.MARKET)}
        r = await self.session.post(STEAM_URL.MARKET / "cancelbuyorder/", data=data, headers=headers)
        rj = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Failed to cancel buy order [{order_id}].", rj)

    async def get_my_listings(self: "SteamClient", *, page_size=100) -> MY_LISTINGS:
        """
        Fetch users market listings.

        :param page_size: listings per page. Steam do not accept greater than 100
        :return: active listings, listings to confirm, buy orders
        :raises ApiError:
        :raises SessionExpired:
        """

        url = STEAM_URL.MARKET / "mylistings"
        params = {"norender": 1, "start": 0, "count": page_size}
        active = []
        to_confirm = []
        buy_orders = []
        classes_map = {}  # ident code: ... , shared classes within whole listings

        # pagination only for active listings. Don't know how to be with others
        more_listings = True
        while more_listings:
            data = await self._fetch_listings(url, params)
            if data["num_active_listings"] > data["pagesize"]:
                more_listings = True
                params["start"] += data["pagesize"]
                params["count"] = data["pagesize"]
            else:
                more_listings = False

            self._parse_item_classes_for_listings(data["assets"], classes_map)
            active.extend(self._parse_listings(data["listings"], classes_map))
            to_confirm.extend(self._parse_listings(data["listings_to_confirm"], classes_map))
            buy_orders.extend(self._parse_buy_orders(data["buy_orders"], classes_map))

        return active, to_confirm, buy_orders

    async def _fetch_listings(self: "SteamClient", url: URL, params: dict) -> dict[str, ...]:
        try:
            r = await self.session.get(url, params=params)
        except ClientResponseError as e:
            raise SessionExpired if e.status == 400 else e

        rj: dict[str, ...] = await r.json()
        if not rj.get("success"):
            raise ApiError("Failed to fetch user listings.", rj)

        return rj

    @classmethod
    def _parse_item_classes_for_listings(
        cls: Type["SteamClient"],
        assets: dict[str, dict[str, dict[str, dict[str, ...]]]],
        classes_map: dict[str, ItemClass],
    ):
        for app_id, app_data in assets.items():
            for context_id, context_data in app_data.items():
                for a_data in context_data.values():
                    key = create_ident_code(a_data["classid"], app_id)
                    if key not in classes_map:
                        classes_map[key] = cls._create_item_class_from_data(a_data, (a_data,))  # some trick

    def _parse_listings(
        self: "SteamClient",
        listings: list[dict[str, ...]],
        classes_map: dict[str, ItemClass],
    ) -> tuple[MyMarketListing, ...]:
        return tuple(
            MyMarketListing(
                id=int(l_data["listingid"]),
                price=l_data["price"] / 100,
                lister_steam_id=self.steam_id,
                time_created=datetime.fromtimestamp(l_data["time_created"]),
                item=MarketListingItem(
                    id=int(l_data["asset"]["id"]),
                    unowned_id=int(l_data["asset"]["unowned_id"]) if "unowned_id" in l_data["asset"] else None,
                    owner_id=self.steam_id,
                    market_id=int(l_data["listingid"]),
                    unowned_context_id=int(l_data["asset"]["unowned_contextid"])
                    if "unowned_contextid" in l_data["asset"]
                    else None,
                    class_=classes_map[create_ident_code(l_data["asset"]["classid"], l_data["asset"]["appid"])],
                    amount=int(l_data["asset"]["amount"]),
                ),
                status=MarketListingStatus(l_data["status"]),
                active=bool(l_data["active"]),
                item_expired=l_data["item_expired"],
                cancel_reason=l_data["cancel_reason"],
                time_finish_hold=l_data["time_finish_hold"],
            )
            for l_data in listings
        )

    @classmethod
    def _parse_buy_orders(
        cls: Type["SteamClient"],
        orders: list[dict[str, ...]],
        classes_map: dict[str, ItemClass],
    ) -> list[BuyOrder, ...]:
        orders_list = []
        for o_data in orders:
            class_ident_key = create_ident_code(o_data["description"]["classid"], o_data["description"]["appid"])
            if class_ident_key not in classes_map:
                classes_map[class_ident_key] = cls._create_item_class_from_data(
                    o_data["description"], (o_data["description"],)
                )  # same some trick
            orders_list.append(
                BuyOrder(
                    id=int(o_data["buy_orderid"]),
                    price=int(o_data["price"]) / 100,
                    item_class=classes_map[class_ident_key],
                    quantity=int(o_data["quantity"]),
                    quantity_remaining=int(o_data["quantity_remaining"]),
                )
            )

        return orders_list

    @overload
    async def buy_market_listing(
        self,
        listing: MarketListing,
    ) -> float:
        ...

    @overload
    async def buy_market_listing(
        self,
        listing: int,
        price: float,
        market_hash_name: str,
        game: GameType,
        *,
        fee: int = ...,
    ) -> float:
        ...

    async def buy_market_listing(
        self: "SteamClient",
        listing: int | MarketListing,
        price: float = None,
        market_hash_name: str = None,
        game: GameType = None,
        *,
        fee: float = None,
    ) -> float:
        """
        Buy item listing from market.
        Unfortunately, Steam requires referer header to buy item,
        so `market hash name` and `game` is mandatory args.

        .. note:: Make sure that listing converted currency is wallet currency!

        :param listing: id for listing itself (aka market id) or `MarketListing`
        :param price: price in `1.24` format, can be found on listing data in
            Steam under field `converted_price` divided by 100
        :param market_hash_name: as arg name
        :param game: as arg name&type
        :param fee: if fee of listing is different from default one,
            can be found on listing data in Steam under field `converted_fee` divided by 100.
            If you don't know what is this - then you definitely do not need it
        :return: wallet balance
        :raises ApiError: for regular reasons
        :raises ValueError:
        """

        if isinstance(listing, MarketListing):
            if listing.converted_currency is self.currency:
                raise ValueError(
                    f"Currency of listing [{listing.converted_currency}] is "
                    f"different from wallet [{self.currency}] one!"
                )

            listing_id = listing.id
            price = listing.converted_price
            fee = listing.converted_fee
            market_hash_name = listing.item.class_.market_hash_name
            game = listing.item.class_.game
        else:
            listing_id = listing

        price = int(price * 100)
        if fee is None:
            steam_fee = floor(price * self._steam_fee) or 1
            publ_fee = floor(price * self._publisher_fee) or 1
            fee = steam_fee + publ_fee
        else:
            fee = int(fee * 100)

        total = price + fee
        data = {
            "sessionid": self.session_id,
            "currency": self._wallet_currency.value,
            "subtotal": price,
            "fee": fee,
            "total": total,
            "quantity": 1,
        }
        headers = {"Referer": str(STEAM_URL.MARKET / f"listings/{game[0]}/{market_hash_name}")}
        r = await self.session.post(STEAM_URL.MARKET / f"buylisting/{listing_id}", data=data, headers=headers)
        rj: dict[str, dict[str, str]] = await r.json()
        if not rj.get("wallet_info", {}).get("success"):
            raise ApiError(
                f"Failed to buy listing [{listing_id}] of `{market_hash_name}` "
                f"for {total / 100} {self._wallet_currency.name}.",
                rj,
            )

        return int(rj["wallet_info"]["wallet_balance"]) / 100

    async def get_my_market_history(
        self: "SteamClient",
        *,
        predicate: PREDICATE = None,
        page_size=500,
    ) -> list[MarketHistoryEvent]:
        url = STEAM_URL.MARKET / "myhistory"
        params = {"norender": 1, "start": 0, "count": page_size}
        events = []
        classes_map = {}  # ident code: ... , shared classes within whole listings
        assets_map = {}  # ident code: ...
        listings_map = {}  # listings id / listing_purchase id

        more_listings = True
        while more_listings:
            data = await self._fetch_listings(url, params)
            if data["total_count"] > data["pagesize"]:
                more_listings = True
                params["start"] += data["pagesize"]
                params["count"] = data["pagesize"]
            else:
                more_listings = False

            self._parse_item_classes_for_listings(data["assets"], classes_map)
            self._parse_assets_for_history_listings(data["assets"], classes_map, assets_map)
            self._parse_history_listings(data, assets_map, listings_map)
            events.extend(self._parse_history_events(data, listings_map))

        return list(e for e in events if predicate(e)) if predicate else events

    @staticmethod
    def _parse_assets_for_history_listings(
        data: dict[str, dict[str, dict[str, dict[str, ...]]]],
        classes_map: dict[str, ItemClass],
        assets_map: dict[str, MarketHistoryListingItem],
    ):
        for app_id, app_data in data.items():
            for context_id, context_data in app_data.items():
                for a_data in context_data.values():
                    key = create_ident_code(a_data["id"], app_id, context_id)
                    if key not in assets_map:
                        assets_map[key] = MarketHistoryListingItem(
                            id=int(a_data["id"]),
                            class_=classes_map[create_ident_code(a_data["classid"], app_id)],
                            unowned_id=int(a_data["unowned_id"]),
                            unowned_context_id=int(a_data["unowned_contextid"]),
                            rollback_new_id=int(a_data["rollback_new_id"]) if "rollback_new_id" in a_data else None,
                            rollback_new_context_id=int(a_data["rollback_new_contextid"])
                            if "rollback_new_contextid" in a_data
                            else None,
                        )

    @staticmethod
    def _parse_history_listings(
        data: dict[str, dict[str, dict[str, ...]]],
        assets_map: dict[str, MarketHistoryListingItem],
        listings_map: dict[str, MarketHistoryListing],
    ):
        for l_id, l_data in data["listings"].items():
            if l_id not in listings_map:
                listings_map[l_id] = MarketHistoryListing(
                    id=int(l_data["listingid"]),
                    price=int(l_data["price"]) / 100,
                    item=assets_map[
                        create_ident_code(
                            l_data["asset"]["id"],
                            l_data["asset"]["appid"],
                            l_data["asset"]["contextid"],
                        )
                    ],
                    original_price=int(l_data["original_price"]) / 100,
                    cancel_reason=l_data.get("cancel_reason"),
                )

        for p_id, p_data in data["purchases"].items():
            if p_id not in listings_map:
                listing = MarketHistoryListing(
                    id=int(p_data["listingid"]),
                    item=assets_map[
                        create_ident_code(
                            p_data["asset"]["id"],
                            p_data["asset"]["appid"],
                            p_data["asset"]["contextid"],
                        )
                    ],
                    purchase_id=int(p_data["purchaseid"]),
                    steamid_purchaser=int(p_data["steamid_purchaser"]),
                    received_amount=int(p_data["received_amount"]) / 100,
                )
                listing.item.new_id = int(p_data["asset"]["new_id"])
                listing.item.new_context_id = int(p_data["asset"]["new_contextid"])

                listings_map[p_id] = listing

    @staticmethod
    def _parse_history_events(
        data: dict[str, list[dict[str, ...]] | dict[str, dict[str, ...]]],
        listings_map: dict[str, MarketHistoryListing],
    ) -> list[MarketHistoryEvent]:
        events = []
        for e_data in data["events"]:
            listing_key = e_data["listingid"]
            if "purchaseid" in e_data:
                listing_key += "_" + e_data["purchaseid"]

            events.append(
                MarketHistoryEvent(
                    listing=listings_map[listing_key],
                    time_event=datetime.fromtimestamp(e_data["time_event"]),
                    type=MarketHistoryEventType(e_data["event_type"]),
                )
            )

        return events

    @overload
    async def fetch_price_history(self, obj: EconItem) -> list[PriceHistoryEntry]:
        ...

    @overload
    async def fetch_price_history(self, obj: ItemClass) -> list[PriceHistoryEntry]:
        ...

    @overload
    async def fetch_price_history(self, obj: str, app_id: int) -> list[PriceHistoryEntry]:
        ...

    async def fetch_price_history(self: "SteamClient", obj: str, app_id: int = None) -> list[PriceHistoryEntry]:
        """
        Fetch price history.
        Prices always will be same currency as a wallet.

        .. seealso:: https://github.com/Revadike/InternalSteamWebAPI/wiki/Get-Market-Price-History

        .. warning:: This request is rate limited by Steam.

        :param obj: `EconItem` or `ItemClass` or market hash name
        :param app_id:
        :return: list of `PriceHistoryEntry`
        :raises ApiError:
        """

        if isinstance(obj, EconItem):
            name = obj.class_.market_hash_name
            app_id = obj.class_.game.app_id
        elif isinstance(obj, ItemClass):
            name = obj.market_hash_name
            app_id = obj.game.app_id
        else:  # str
            name = obj

        params = {"appid": app_id, "market_hash_name": name}
        r = await self.session.get(STEAM_URL.MARKET / "pricehistory", params=params)
        rj: dict[str, list[list]] = await r.json()
        if not rj.get("success"):
            raise ApiError(f"Failed to fetch `{name}` price history.", rj)

        return [
            PriceHistoryEntry(
                date=datetime.strptime(e_data[0], PRICE_HISTORY_ENTRY_DATE_FORMAT),
                price=e_data[1],
                daily_volume=int(e_data[2]),
            )
            for e_data in rj["prices"]
        ]



"""
This is `aiosteampy` package.
"""

from .exceptions import ApiError, LoginError, ConfirmationError, CaptchaRequired, SessionExpired
from .constants import Game, STEAM_URL, Currency, Language, TradeOfferStatus, MarketListingStatus
from .client import SteamClient, SteamPublicClient



from re import compile
from urllib.parse import quote
from json import loads
from http.cookies import SimpleCookie

from aiohttp import ClientSession
from aiohttp.client import _RequestContextManager
from yarl import URL

from .guard import SteamGuardMixin
from .confirmation import ConfirmationMixin
from .login import LoginMixin
from .trade import TradeMixin
from .market import MarketMixin
from .public import SteamPublicMixin, INV_PAGE_SIZE, PREDICATE, PRIVATE_USER_EXC_MSG

from .models import Notifications, EconItem
from .constants import STEAM_URL, Currency, GameType, Language, CORO
from .exceptions import ApiError, SessionExpired
from .utils import get_cookie_value_from_session, steam_id_to_account_id, account_id_to_steam_id

__all__ = ("SteamClient", "SteamPublicClient", "SteamCommunityMixin")

API_KEY_RE = compile(r"<p>Key: (?P<api_key>[0-9A-F]+)</p>")
WALLET_INFO_RE = compile(r"g_rgWalletInfo = (?P<info>.+);")
TRADE_TOKEN_RE = compile(r"\d+&token=(?P<token>.+)\" readonly")

API_KEY_CHECK_STR = "<h2>Access Denied</h2>"
API_KEY_CHECK_STR1 = "You must have a validated email address to create a Steam Web API key"
STEAM_LANG_COOKIE = "Steam_Language"
DEF_COUNTRY = "UA"


class SteamCommunityMixin(SteamGuardMixin, ConfirmationMixin, LoginMixin, MarketMixin, TradeMixin, SteamPublicMixin):
    """
    Mixin class, but with `__init__`. Inherits all other mixins.
    Need in case you want to use multiple inheritance with `__slots__`.
    """

    __slots__ = ()

    session: ClientSession
    username: str
    steam_id: int
    trade_token: str

    def __init__(
        self,
        username: str,
        password: str,
        steam_id: int,
        *,
        shared_secret: str,
        identity_secret: str,
        api_key: str = None,
        trade_token: str = None,
        wallet_currency: Currency = None,
        wallet_country=DEF_COUNTRY,
        steam_fee=0.05,
        publisher_fee=0.1,
        lang=Language.ENGLISH,
        tz_offset=0.0,
        session: ClientSession = None,
    ):

        self.session = session or ClientSession(raise_for_status=True)
        self.username = username
        self.steam_id = account_id_to_steam_id(steam_id) if steam_id < 4294967296 else steam_id  # steam id64
        self.trade_token = trade_token

        self._password = password
        self._shared_secret = shared_secret
        self._identity_secret = identity_secret
        self._api_key = api_key
        self._steam_fee = steam_fee
        self._publisher_fee = publisher_fee
        self._wallet_currency = wallet_currency

        self._wallet_country = wallet_country

        super().__init__()

        self._set_init_cookies(lang, tz_offset)

    @property
    def account_id(self) -> int:
        """Steam id32."""
        return steam_id_to_account_id(self.steam_id)

    @property
    def trade_url(self) -> URL:
        return STEAM_URL.TRADE / "new/" % {"partner": self.account_id, "token": self.trade_token or ""}

    @property
    def profile_url(self) -> URL:
        return STEAM_URL.COMMUNITY / f"profiles/{self.steam_id}"

    @property
    def language(self) -> Language:
        """Language of Steam html pages, json info, descriptions, etc."""
        return Language(get_cookie_value_from_session(self.session, STEAM_URL.STORE.host, STEAM_LANG_COOKIE))

    @property
    def country(self) -> str:
        """Just wallet country. Needed for public methods."""
        return self._wallet_country

    @property
    def currency(self) -> Currency:
        """Alias for wallet currency. Needed for public methods."""
        return self._wallet_currency

    def _set_init_cookies(self, lang: str, tz_offset: float):
        urls = (STEAM_URL.COMMUNITY, STEAM_URL.STORE, STEAM_URL.HELP)
        for url in urls:
            c = SimpleCookie()
            c[STEAM_LANG_COOKIE] = lang
            c[STEAM_LANG_COOKIE]["path"] = "/"
            c[STEAM_LANG_COOKIE]["domain"] = url.host
            c[STEAM_LANG_COOKIE]["secure"] = True

            self.session.cookie_jar.update_cookies(cookies=c, response_url=url)

        c_key = "timezoneOffset"
        for url in urls:
            c = SimpleCookie()
            c[c_key] = str(tz_offset).replace(".", ",")
            c[c_key]["path"] = "/"
            c[c_key]["domain"] = url.host
            c[c_key]["SameSite"] = True

            self.session.cookie_jar.update_cookies(cookies=c, response_url=url)

    async def _init_data(self):
        if not self._api_key:
            self._api_key = await self._fetch_api_key()
            not self._api_key and await self.register_new_api_key()

        if not self.trade_token:
            self.trade_token = await self._fetch_trade_token()
            not self.trade_token and await self.register_new_trade_url()

        if not self._steam_fee or not self._publisher_fee or not self._wallet_currency:
            wallet_info = await self._fetch_wallet_info()
            self._wallet_country = wallet_info["wallet_country"]
            if not self._steam_fee:
                self._steam_fee = float(wallet_info["wallet_fee_percent"])
            if not self._publisher_fee:
                self._publisher_fee = float(wallet_info["wallet_publisher_fee_percent_default"])
            if not self._wallet_currency:
                self._wallet_currency = Currency(wallet_info["wallet_currency"])

    async def _fetch_wallet_info(self) -> dict[str, str | int]:
        # fetching inventory may reset new items notifs count
        r = await self.session.get(self.profile_url / "inventory", headers={"Referer": str(self.profile_url)})
        rt = await r.text()
        info: dict = loads(WALLET_INFO_RE.search(rt)["info"])
        if not info.get("success"):
            raise ApiError("Failed to fetch wallet info from inventory.", info)

        return info

    async def get_wallet_balance(self) -> tuple[float, Currency]:
        """
        Fetch wallet balance and currency.

        :return: tuple of balance and `Currency`
        """

        r = await self.session.get(STEAM_URL.STORE / "api/getfundwalletinfo")
        rj = await r.json()
        if not rj.get("success"):
            raise ApiError("Failed to fetch wallet info.", rj)

        self._wallet_currency = Currency.by_name(rj["user_wallet"]["currency"])
        return int(rj["user_wallet"]["amount"]) / 100, self._wallet_currency

    async def register_new_trade_url(self) -> URL:
        """
        Register new trade url. Cache token.

        :return: trade url
        """

        r = await self.session.post(
            self.profile_url / "tradeoffers/newtradeurl",
            data={"sessionid": self.session_id},
        )
        token: str = await r.json()

        self.trade_token = quote(token, safe="~()*!.'")  # https://stackoverflow.com/a/72449666/19419998
        return self.trade_url

    async def _fetch_trade_token(self) -> str | None:
        r = await self.session.get(self.profile_url / "tradeoffers/privacy")
        rt = await r.text()

        search = TRADE_TOKEN_RE.search(rt)
        return search["token"] if search else None

    async def _fetch_api_key(self) -> str | None:
        r = await self.session.get(STEAM_URL.COMMUNITY / "dev/apikey")
        rt = await r.text()
        if API_KEY_CHECK_STR in rt or API_KEY_CHECK_STR1 in rt:
            raise ApiError(API_KEY_CHECK_STR1)

        search = API_KEY_RE.search(rt)
        return search["api_key"] if search else None

    async def register_new_api_key(self, domain="https://github.com/somespecialone/aiosteampy") -> str:
        """
        Register new api key, cache it and return.

        :param domain: On which domain api key will be registered.
            Default - "https://github.com/somespecialone/aiosteampy"
        :return: api key
        """

        data = {
            "domain": domain,
            "agreeToTerms": "agreed",
            "sessionid": self.session_id,
            "Submit": "Register",
        }
        r = await self.session.post(STEAM_URL.COMMUNITY / "dev/registerkey", data=data)
        rt = await r.text()

        self._api_key = API_KEY_RE.search(rt)["api_key"]
        return self._api_key

    async def get_notifications(self) -> Notifications:
        """Get notifications count."""

        headers = {"Referer": str(self.profile_url)}
        r = await self.session.get(STEAM_URL.COMMUNITY / "actions/GetNotificationCounts", headers=headers)
        rj = await r.json()
        return Notifications(*(rj["notifications"][str(i)] for i in range(1, 12) if i != 7))

    # TODO check primary events resetting conditions.
    def reset_items_notifications(self) -> CORO[_RequestContextManager]:
        """
        Fetching your inventory page, which resets new items notifications to 0.

        .. seealso:: https://github.com/DoctorMcKay/node-steamcommunity/blob/851c14bd93008579e7a308ea8ecda873996baa1f/index.js#L405

        """

        return self.session.get(self.profile_url / "inventory/")

    async def get_inventory(
        self,
        game: GameType,
        *,
        predicate: PREDICATE = None,
        page_size=INV_PAGE_SIZE,
    ) -> list[EconItem]:
        """
        Fetches self inventory.

        :param game: just Steam Game
        :param page_size: max items on page. Current Steam limit is 2000
        :param predicate: callable with single arg `EconItem`, must return bool
        :return: list of `EconItem`
        :raises ApiError: for ordinary reasons
        :raises SessionExpired:
        """

        try:
            inv = await self.get_user_inventory(self.steam_id, game, predicate=predicate, page_size=page_size)
        except ApiError as e:
            raise SessionExpired if e.msg == PRIVATE_USER_EXC_MSG else e  # self inventory can't be private
            # may I need to cache client items?
        return inv


class SteamClient(SteamCommunityMixin):
    """Base class in hierarchy ..."""

    __slots__ = (
        "_is_logged",
        "session",
        "username",
        "steam_id",
        "_password",
        "_shared_secret",
        "_identity_secret",
        "_api_key",
        "trade_token",
        "_device_id",
        "_confirmation_storage",
        "_steam_fee",
        "_publisher_fee",
        "_wallet_currency",
        "_wallet_country",
        "_trades_storage",
    )


class SteamPublicClient(SteamPublicMixin):
    """Class for public methods that not requires login."""

    __slots__ = ("session", "language", "currency", "country")

    def __init__(
        self,
        *,
        language=Language.ENGLISH,
        currency=Currency.USD,
        country=DEF_COUNTRY,
        session: ClientSession = None,
    ):
        self.session = session or ClientSession(raise_for_status=True)

        self.language = language
        self.currency = currency
        self.country = country



from typing import TYPE_CHECKING, overload, Literal, Type, TypeAlias, Callable
from datetime import datetime
from json import dumps as jdumps

from yarl import URL

from .exceptions import ApiError
from .constants import STEAM_URL, CORO
from .models import (
    TradeOffer,
    TradeOfferItem,
    TradeOfferStatus,
    ItemClass,
    HistoryTradeOffer,
    HistoryTradeOfferItem,
    EconItem,
    EconItemType,
)
from .utils import create_ident_code, to_int_boolean, steam_id_to_account_id, account_id_to_steam_id
from .typed import TradeOffersSummary

if TYPE_CHECKING:
    from .client import SteamClient

__all__ = ("TradeMixin",)

TRADE_OFFERS: TypeAlias = tuple[list[TradeOffer], list[TradeOffer]]
ERROR_MSG = "You can't accept your offer! Are you trying to cancel outgoing offer?"
PRED: TypeAlias = Callable[[TradeOffer], bool]
T_KWARGS: TypeAlias = int | str | float


class TradeMixin:
    """
    Mixin with trade offers related methods.
    Depends on :class:`aiosteampy.confirmation.ConfirmationMixin`,
    :class:`aiosteampy.public.SteamPublicMixin`.
    """

    __slots__ = ()

    async def remove_trade_offer(self, offer_id: int, offer: TradeOffer | None):
        """
        Remove trade offer silently from cache.

        You can override this method to provide your custom storage.
        """

    async def store_trade_offer(self, offer_id: int, offer: TradeOffer):
        """
        Cache trade offer to inner store.

        You can override this method to provide your custom storage.
        """

    async def store_multiple_trade_offers(self, offer_ids: list[int], offers: list[TradeOffer]):
        """
        Cache multiple trade offers to inner store.

        You can override this method to provide your custom storage.
        """

        for index, offer_id in enumerate(offer_ids):
            await self.store_trade_offer(offer_id, offers[index])

    async def get_trade_offer(self, offer_id: int) -> TradeOffer | None:
        """
        Get trade offer from storage.

        You can override this method to provide your custom storage.
        """

    async def get_trade_offers(self, predicate: PRED = None) -> list[TradeOffer]:
        """
        Cached trade offers.

        You can override this method to provide your custom storage.
        """

    async def get_or_fetch_trade_offer(self, offer_id: int) -> TradeOffer:
        """
        Get specific trade from cache. Fetch it and store, if there is no one.

        :raises ApiError: if there is error when trying to fetch `TradeOffer`
        """

        trade = await self.get_trade_offer(offer_id)
        if not trade:
            trade = await self.fetch_trade(offer_id)

        return trade

    async def fetch_trade(self: "SteamClient", offer_id: int) -> TradeOffer:
        """
        Fetch trade offer from Steam.

        :raises ApiError:
        """

        params = {
            "key": self._api_key,
            "tradeofferid": offer_id,
            "language": self.language,
            "get_descriptions": 1,
        }
        r = await self.session.get(STEAM_URL.API.IEconService.GetTradeOffer, params=params)
        rj: dict[str, ...] = await r.json()
        if not rj.get("response"):
            raise ApiError(f"Can't fetch trade offer [{offer_id}].", rj)

        data: dict[str, dict[str, ...] | list[dict[str, ...]]] = rj["response"]
        classes_map = {}
        self._update_classes_map_for_trades(data["descriptions"], classes_map)

        trade = self._create_trade_offer_from_data(data["offer"], classes_map)
        await self.store_trade_offer(trade.id, trade)
        return trade

    @classmethod
    def _update_classes_map_for_trades(cls: Type["SteamClient"], data: list[dict], classes_map: dict):
        for desc in data:
            key = create_ident_code(desc["classid"], desc["appid"])
            if key not in classes_map:
                classes_map[key] = cls._create_item_class_from_data(desc, (desc,))

    def _create_trade_offer_from_data(
        self: "SteamClient",
        data: dict[str, ...],
        classes_map: dict[str, ItemClass],
    ) -> TradeOffer:
        return TradeOffer(
            id=int(data["tradeofferid"]),
            owner=self.steam_id,
            partner_id=data["accountid_other"],
            is_our_offer=data["is_our_offer"],
            expiration_time=datetime.fromtimestamp(data["expiration_time"]),
            time_created=datetime.fromtimestamp(data["time_created"]),
            time_updated=datetime.fromtimestamp(data["time_updated"]),
            items_to_give=self._parse_assets_for_trade(data.get("items_to_give", ()), classes_map),
            items_to_receive=self._parse_assets_for_trade(data.get("items_to_receive", ()), classes_map),
            message=data["message"],
            status=TradeOfferStatus(data["trade_offer_state"]),
        )

    @classmethod
    def _parse_assets_for_trade(
        cls: Type["SteamClient"],
        items: list[dict[str, ...]],
        classes_map: dict[str, ItemClass],
    ) -> list[TradeOfferItem]:
        return [
            TradeOfferItem(
                id=a_data["assetid"],
                class_id=int(a_data["classid"]),
                game=cls._find_game_for_asset(a_data, (a_data,)),
                amount=int(a_data["amount"]),
                class_=classes_map.get(create_ident_code(a_data["classid"], a_data["appid"])),
                missing=a_data["missing"],
                est_usd=int(a_data["est_usd"]),
            )
            for a_data in items
        ]

    @overload
    async def fetch_trade_offers(
        self,
        *,
        time_historical_cutoff: int = ...,
        sent: bool = ...,
        received: bool = ...,
        **kwargs: T_KWARGS,
    ) -> TRADE_OFFERS:
        ...

    @overload
    async def fetch_trade_offers(
        self,
        *,
        active_only: Literal[False],
        historical_only: Literal[True],
        sent: bool = ...,
        received: bool = ...,
        **kwargs: T_KWARGS,
    ) -> TRADE_OFFERS:
        ...

    async def fetch_trade_offers(
        self: "SteamClient",
        *,
        active_only=True,
        time_historical_cutoff: int = None,
        historical_only=False,
        sent=True,
        received=True,
        **kwargs: T_KWARGS,
    ) -> TRADE_OFFERS:
        """
        Fetch trade offers from Steam Web Api.

        :param active_only: fetch active, changed since `time_historical_cutoff` tradeoffs only or not
        :param time_historical_cutoff: timestamp for `active_only`
        :param historical_only: opposite for `active_only`
        :param sent: include sent offers or not
        :param received: include received offers or not
        :return: sent trades, received trades lists
        """

        params = {
            "key": self._api_key,
            "active_only": to_int_boolean(active_only),
            "get_sent_offers": to_int_boolean(sent),
            "get_received_offers": to_int_boolean(received),
            "historical_only": to_int_boolean(historical_only),
            "get_descriptions": 1,
            "cursor": 0,
            "language": self.language,
            **kwargs,
        }
        if active_only and time_historical_cutoff is not None:
            params["time_historical_cutoff"] = time_historical_cutoff

        classes_map: dict[str, ItemClass] = {}
        offer_sent_datas = []
        offer_received_datas = []

        while True:
            r = await self.session.get(STEAM_URL.API.IEconService.GetTradeOffers, params=params)
            rj = await r.json()
            if not rj.get("response"):
                raise ApiError(f"Can't fetch trade offers.", rj)

            data: dict[str, dict[str, ...] | list[dict[str, ...]]] = rj["response"]
            offer_sent_datas.extend(data.get("trade_offers_sent", ()))
            offer_received_datas.extend(data.get("trade_offers_received", ()))
            self._update_classes_map_for_trades(data["descriptions"], classes_map)

            params["cursor"] = data.get("next_cursor", 0)
            if not params["cursor"]:
                break

        o_sent = [self._create_trade_offer_from_data(d, classes_map) for d in offer_sent_datas]
        o_received = [self._create_trade_offer_from_data(d, classes_map) for d in offer_received_datas]

        await self.store_multiple_trade_offers(
            [*(t.id for t in o_sent), *(t.id for t in o_received)],
            [*o_sent, *o_received],
        )

        return o_sent, o_received

    async def get_trade_offers_summary(self: "SteamClient") -> TradeOffersSummary:
        r = await self.session.get(STEAM_URL.API.IEconService.GetTradeOffersSummary, params={"key": self._api_key})
        rj: dict[str, TradeOffersSummary] = await r.json()
        if not rj.get("response"):
            raise ApiError(f"Can't fetch trade offers summary.", rj)

        return rj["response"]

    async def get_trade_receipt(self: "SteamClient", offer_id: int) -> HistoryTradeOffer:
        """Fetch single trade offer from history."""

        params = {
            "key": self._api_key,
            "tradeid": offer_id,
            "get_descriptions": 1,
            "language": self.language,
        }
        r = await self.session.get(STEAM_URL.API.IEconService.GetTradeStatus, params=params)
        rj = await r.json()
        if not rj.get("response"):
            raise ApiError(f"Can't fetch trade status.", rj)

        classes_map = {}
        self._update_classes_map_for_trades(rj["response"]["descriptions"], classes_map)

        return self._create_history_trade_offer_from_data(rj["response"]["trades"][0], classes_map)

    async def get_trade_history(
        self: "SteamClient",
        max_trades=100,
        *,
        start_after_time=0,
        start_after_trade_id=0,
        navigating_back=False,
        include_failed=True,
        **kwargs: T_KWARGS,
    ) -> tuple[list[HistoryTradeOffer], int]:
        """
        Fetch history trades with changed assets data.
        You can paginate by yourself with this method.

        :param max_trades: page size
        :param start_after_time: timestamp
        :param start_after_trade_id:
        :param navigating_back:
        :param include_failed:
        :return: list of `HistoryTradeOffer`, total trades count
        """

        params = {
            "key": self._api_key,
            "max_trades": max_trades,
            "get_descriptions": 1,
            "include_total": 1,
            "include_failed": to_int_boolean(include_failed),
            "navigating_back": to_int_boolean(navigating_back),
            "start_after_time": start_after_time,
            "start_after_tradeid": start_after_trade_id,
            "language": self.language,
            **kwargs,
        }
        r = await self.session.get(STEAM_URL.API.IEconService.GetTradeHistory, params=params)
        rj = await r.json()
        if not rj.get("response"):
            raise ApiError(f"Can't fetch trades history.", rj)

        classes_map = {}
        data: dict[str, int | bool | dict[str, ...] | list[dict[str, ...]]] = rj["response"]
        self._update_classes_map_for_trades(data["descriptions"], classes_map)

        return (
            [self._create_history_trade_offer_from_data(d, classes_map) for d in data["trades"]],
            data["total_trades"],
        )

    def _create_history_trade_offer_from_data(
        self: "SteamClient",
        data: dict,
        classes_map: dict[str, ItemClass],
    ) -> HistoryTradeOffer:
        return HistoryTradeOffer(
            id=int(data["tradeid"]),
            owner=self.steam_id,
            partner_id=steam_id_to_account_id(int(data["steamid_other"])),
            time_init=datetime.fromtimestamp(data["time_init"]),
            status=TradeOfferStatus(data["status"]),
            assets_given=self._parse_assets_for_history_trades(data.get("assets_given", ()), classes_map),
            assets_received=self._parse_assets_for_history_trades(data.get("assets_received", ()), classes_map),
        )

    @classmethod
    def _parse_assets_for_history_trades(
        cls: Type["SteamClient"],
        items: list[dict[str, ...]],
        classes_map: dict[str, ItemClass],
    ) -> tuple[HistoryTradeOfferItem, ...]:
        return tuple(
            HistoryTradeOfferItem(
                id=int(a_data["assetid"]),
                class_id=int(a_data["classid"]),
                game=cls._find_game_for_asset(a_data, (a_data,)),
                amount=int(a_data["amount"]),
                class_=classes_map.get(create_ident_code(a_data["classid"], a_data["appid"])),
                new_asset_id=int(a_data["new_assetid"]),
                new_context_id=int(a_data["new_contextid"]),
            )
            for a_data in items
        )

    async def _do_action_with_offer(self: "SteamClient", offer_id: int, action: Literal["cancel", "decline"]) -> int:
        r = await self.session.post(STEAM_URL.TRADE / f"{offer_id}/{action}", data={"sessionid": self.session_id})
        rj: dict[str, str] = await r.json()
        return int(rj.get("tradeofferid", 0))

    async def cancel_trade_offer(self: "SteamClient", offer: int | TradeOffer):
        """Cancel outgoing trade offer. Remove offer from cache."""

        if isinstance(offer, TradeOffer):
            if not offer.is_our_offer:
                raise ValueError("You can't cancel not your offer! Are you trying to decline incoming offer?")
            offer_id = offer.id
            to_remove = offer
        else:
            offer_id = offer
            to_remove = None
        resp_offer_id = await self._do_action_with_offer(offer_id, "cancel")
        if not resp_offer_id or resp_offer_id != offer_id:
            raise ApiError(f"Error while try to cancel offer [{offer_id}].")

        await self.remove_trade_offer(offer_id, to_remove)  # remove after making sure that all is okay

    async def decline_trade_offer(self: "SteamClient", offer: int | TradeOffer):
        """Decline incoming trade offer. Remove offer from cache."""

        if isinstance(offer, TradeOffer):
            if offer.is_our_offer:
                raise ValueError("You can't decline your offer! Are you trying to cancel outgoing offer?")
            offer_id = offer.id
            to_remove = offer
        else:
            offer_id = offer
            to_remove = None
        resp_offer_id = await self._do_action_with_offer(offer_id, "decline")
        if not resp_offer_id or resp_offer_id != offer_id:
            raise ApiError(f"Error while try to decline offer [{offer_id}].")

        await self.remove_trade_offer(offer_id, to_remove)  # remove after making sure that all is okay

    @overload
    async def accept_trade_offer(self, offer: TradeOffer):
        ...

    @overload
    async def accept_trade_offer(self, offer: int, partner: int = ...):
        ...

    async def accept_trade_offer(self: "SteamClient", offer: int | TradeOffer, partner: int = None):
        """
        Accept trade offer, yes. Remove offer from cache.

        .. note::
            Auto confirm accepting if needed.

            If you not pass ``partner`` but pass `trade offer id` -
            fetches :class:`aiosteampy.models.TradeOffer` from `Steam`.

        :param offer: `TradeOffer` or trade offer id
        :param partner: partner account id (id32) or steam id (id64)
        :raises ConfirmationError: if confirmation can't be found or while do action with
        :raises ApiError: if there is error when trying to fetch `TradeOffer`
        """

        if isinstance(offer, TradeOffer):
            if offer.is_our_offer:
                raise ValueError(ERROR_MSG)
            offer_id = offer.id
            partner = offer.partner_id64
            to_remove = TradeOffer
        else:  # int
            if not partner:
                fetched = await self.get_or_fetch_trade_offer(offer)
                return await self.accept_trade_offer(fetched)

            offer_id = offer
            partner = account_id_to_steam_id(partner) if partner < 4294967296 else partner  # 2**32
            to_remove = None

        data = {
            "sessionid": self.session_id,
            "tradeofferid": offer_id,
            "serverid": 1,
            "partner": partner,
            "captcha": "",
        }
        url_base = STEAM_URL.TRADE / str(offer_id)
        r = await self.session.post(url_base / "accept", data=data, headers={"Referer": str(url_base)})
        rj: dict[str, ...] = await r.json()
        if rj.get("needs_mobile_confirmation"):
            await self.confirm_trade_offer(offer_id)

        await self.remove_trade_offer(offer_id, to_remove)

    @overload
    async def make_trade_offer(
        self,
        obj: int,
        to_give: list[EconItem] = ...,
        to_receive: list[EconItem] = ...,
        message: str = ...,
        *,
        token: str = ...,
        confirm: bool = ...,
        countered_id: int = ...,
    ) -> int:
        ...

    @overload
    async def make_trade_offer(
        self,
        obj: str,
        to_give: list[EconItem] = ...,
        to_receive: list[EconItem] = ...,
        message: str = ...,
        *,
        confirm: bool = ...,
        countered_id: int = ...,
    ) -> int:
        ...

    async def make_trade_offer(
        self: "SteamClient",
        obj: int | str,
        to_give: list[EconItemType] = (),
        to_receive: list[EconItemType] = (),
        message="",
        *,
        token: str = None,
        confirm=True,
        countered_id: int = None,
        **kwargs: T_KWARGS,
    ) -> int:
        """
        Make (send) steam trade offer to partner.

        .. note:: Make sure that partner is in friends list if you not pass trade url or trade token.

        :param obj: partner trade url, partner id(id32 or id64)
        :param token:
        :param to_give:
        :param to_receive:
        :param message:
        :param confirm:
        :param countered_id:
        :param kwargs:
        :return: trade offer id
        :raises ValueError: trade is empty
        """

        if not to_give and not to_receive:
            raise ValueError("You can't make empty trade offer!")

        trade_url, partner_id32, partner_id64, token = self._parse_make_offer_args(obj, token)
        base_url = STEAM_URL.TRADE / "new/"
        referer = trade_url or base_url % {"partner": partner_id32}
        offer_params = {}
        if token:
            referer %= {"token": token}
            offer_params["trade_offer_access_token"] = token
        data = {
            "sessionid": self.session_id,
            "serverid": 1,
            "partner": partner_id64,
            "tradeoffermessage": message,
            "json_tradeoffer": jdumps(
                {
                    "newversion": True,
                    "version": len(to_give) + len(to_receive) + 1,
                    "me": {"assets": [self._to_asset_dict(i) for i in to_give], "currency": [], "ready": False},
                    "them": {"assets": [self._to_asset_dict(i) for i in to_receive], "currency": [], "ready": False},
                }
            ),
            "captcha": "",
            "trade_offer_create_params": jdumps(offer_params),
            **kwargs,
        }
        if countered_id:
            data["tradeofferid_countered"] = countered_id

        headers = {"Referer": str(referer)}
        r = await self.session.post(base_url / "send", data=data, headers=headers)
        rj = await r.json()
        offer_id = int(rj["tradeofferid"])
        if confirm and rj.get("needs_mobile_confirmation"):
            offer_id = await self.confirm_trade_offer(offer_id)

        return offer_id

    @staticmethod
    def _to_asset_dict(obj: EconItemType) -> dict[str, int | str]:
        return {"appid": obj[0], "contextid": str(obj[1]), "amount": obj[2], "assetid": str(obj[3])}

    @staticmethod
    def _parse_make_offer_args(obj: str | int, token: str | None) -> tuple[str | None, int, int, str | None]:
        trade_url = None
        if isinstance(obj, str):  # trade url
            trade_url = URL(obj)
            partner_id32 = int(trade_url.query["partner"])
            partner_id64 = account_id_to_steam_id(partner_id32)
            token = trade_url.query["token"]
            trade_url = str(trade_url)
        else:
            partner_id = obj
            if partner_id < 4294967296:  # 32
                partner_id32 = partner_id
                partner_id64 = account_id_to_steam_id(partner_id)
            else:  # 64
                partner_id64 = partner_id
                partner_id32 = steam_id_to_account_id(partner_id64)

        return trade_url, partner_id32, partner_id64, token

    @overload
    async def counter_trade_offer(
        self,
        obj: TradeOffer,
        to_give: list[EconItemType] = (),
        to_receive: list[EconItemType] = (),
        message="",
        *,
        confirm: bool = ...,
    ) -> int:
        ...

    @overload
    async def counter_trade_offer(
        self,
        obj: int,
        to_give: list[EconItemType] = (),
        to_receive: list[EconItemType] = (),
        message="",
        *,
        partner_id: int,
        confirm: bool = ...,
    ) -> int:
        ...

    def counter_trade_offer(
        self,
        obj: TradeOffer | int,
        to_give: list[EconItemType] = (),
        to_receive: list[EconItemType] = (),
        message="",
        *,
        partner_id: int = None,
        confirm=True,
    ) -> CORO[int]:
        """
        Counter trade offer with another.

        :param obj: `TradeOffer` or trade offer id of which you want to counter
        :param to_give:
        :param to_receive:
        :param message:
        :param partner_id:
        :param confirm:
        :return:
        """

        if isinstance(obj, TradeOffer):
            offer_id = obj.id
            to_give_updated = [*obj.items_to_give, *to_give]
            to_receive_updated = [*obj.items_to_receive, *to_receive]
            if obj.is_our_offer:
                raise ValueError("You can't counter yours offer!")
        else:  # trade id
            offer_id = obj
            to_give_updated = to_give
            to_receive_updated = to_receive

        return self.make_trade_offer(
            partner_id,
            to_give_updated,
            to_receive_updated,
            message,
            confirm=confirm,
            countered_id=offer_id,
        )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['aiosteampy']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.8.3,<4.0.0', 'beautifulsoup4>=4.10.0,<5.0.0', 'rsa>=4.9,<5.0']

setup_kwargs = {
    'name': 'aiosteampy',
    'version': '0.2.1',
    'description': 'Simple library to trade and interact with steam market, webapi, guard',
    'long_description': "<!--header-start-->\n\n# AIOSTEAMPY\n\n[![Made in Ukraine](https://img.shields.io/badge/made_in-ukraine-ffd700.svg?labelColor=0057b7)](https://stand-with-ukraine.pp.ua)\n[![steam](https://shields.io/badge/steam-1b2838?logo=steam)](https://store.steampowered.com/)\n[![license](https://img.shields.io/github/license/somespecialone/aiosteampy)](https://github.com/somespecialone/aiosteampy/blob/master/LICENSE)\n[![pypi](https://img.shields.io/pypi/v/aiosteampy)](https://pypi.org/project/aiosteampy)\n[![versions](https://img.shields.io/pypi/pyversions/aiosteampy)](https://pypi.org/project/aiosteampy)\n[![Tests](https://github.com/somespecialone/aiosteampy/actions/workflows/tests.yml/badge.svg)](https://github.com/somespecialone/aiosteampy/actions/workflows/tests.yml)\n[![Publish](https://github.com/somespecialone/aiosteampy/actions/workflows/publish.yml/badge.svg)](https://github.com/somespecialone/aiosteampy/actions/workflows/publish.yml)\n[![Docs](https://github.com/somespecialone/aiosteampy/actions/workflows/docs.yml/badge.svg)](https://github.com/somespecialone/aiosteampy/actions/workflows/docs.yml)\n[![codecov](https://codecov.io/gh/somespecialone/aiosteampy/branch/master/graph/badge.svg?token=SP7EQKPIQ3)](https://codecov.io/gh/somespecialone/aiosteampy)\n[![CodeFactor](https://www.codefactor.io/repository/github/somespecialone/aiosteampy/badge)](https://www.codefactor.io/repository/github/somespecialone/aiosteampy)\n[![black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n\nPreviously this library was a soft fork of [bukson/steampy](https://github.com/bukson/steampy) and created only to\nprovide asynchronous methods and proxies support.\nBut now it standalone project. Created by myself for steam trading purposes mostly.\n\n---\n\n[![Stand With Ukraine](https://raw.githubusercontent.com/vshymanskyy/StandWithUkraine/main/banner-direct-single.svg)](https://stand-with-ukraine.pp.ua)\n\n<!--header-end-->\n\n---\n\n> **See full documentation [here](https://somespecialone.github.io/aiosteampy) 📖**\n\n---\n\n<!--install-start-->\n\n## Installation\n\npip\n\n```shell\npip install aiosteampy\n```\n\npipenv\n\n```shell\npipenv install aiosteampy\n```\n\npoetry\n\n```shell\npoetry add aiosteampy\n```\n\n<!--install-end-->\n\nAlso, [aiohttp docs](https://docs.aiohttp.org/en/stable/#library-installation) recommends install speedups (`aiodns`\n, `cchardet`, ...)\n\n> Old version still available on PyPI: [asyncsteampy](https://pypi.org/project/asyncsteampy/)\n\n---\n\n<!--intro-start-->\n\nAIOSTEAMPY use [aiohttp](https://github.com/aio-libs/aiohttp) underneath to get asynchronous requests to steam servers,\nwith modern (not really for current moment 😊) async/await syntax.\nProject is similar to [Gobot1234/steam.py](https://github.com/Gobot1234/steam.py) for first look, but with some\ndifferences.\nIt uses only requests and steam apis (documented and not), while `steam.py` implement stateful steam client based on\nwebsocket\nprotobuf, same as [DoctorMcKay/node-steam-user](https://github.com/DoctorMcKay/node-steam-user).\n\n> Generally, project inspired most\n> by [DoctorMcKay/node-steamcommunity](https://github.com/DoctorMcKay/node-steamcommunity)\n> but created with additions and differences, of course.\n\n## Key features\n\n* Stateless: the main idea was a low-middle layer API wrapper of some steam services and methods like market,\n  tradeoffers, confirmations, steamguard, etc. But if you want to cache your entities data (listings, confirmations,\n  ...) there is some methods to help.\n* Declarative: there is models almost for every data.\n* Typed: for editor support most things are typed.\n* Short: I really tried to fit most important for steam trading methods.\n\n## What can I do with this\n\n* Operate with steam trade offers for any manner.\n* Sell, buy items on market. Place, cancel orders.\n* Login trough steam to 3rd party sites.\n* Fetch data from market.\n* Manipulate many accounts with proxies for each session.\n* Store and load cookies to stay logged in.\n\n## What I can't do\n\n* Chat (at least for now).\n* Get apps, packages.\n* All, that need connection to CM.\n* Interact with game servers (inspect CSGO items, ...).\n* Edit profile, social interaction(groups, clans).\n* Handle entities(listings, items, tradeoffers) lifecycle for easy if you need to store it.\n\n<!--intro-end-->\n\n---\n\n## Tests 🧪\n\n**Read [test documentation](https://somespecialone.github.io/aiosteampy/tests/) 📖**\n\n---\n\n<!--footer-start-->\n\n## Contribution 💛\n\nThere is no rules or requirements to contribute. Feedbacks, suggests, other are welcome.\nI will be very grateful for helping me get the things right.\n\n## Credits\n\n* [DoctorMcKay/node-steamcommunity](https://github.com/DoctorMcKay/node-steamcommunity)\n* [Identifying Steam items](https://dev.doctormckay.com/topic/332-identifying-steam-items/)\n* [Revadike/InternalSteamWebAPI](https://github.com/Revadike/InternalSteamWebAPI)\n* [Gobot1234/steam.py](https://github.com/Gobot1234/steam.py)\n* [somespecialone/steam-item-name-ids](https://github.com/somespecialone/steam-item-name-ids)\n* [steamapi.xpaw.me](https://steamapi.xpaw.me/)\n\n<!--footer-end-->\n",
    'author': 'Dmytro Tkachenko',
    'author_email': 'tkachenkodmitriy@yahoo.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/somespecialone/aiosteampy/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



"""Constants and enums, some types"""

from typing import TypeAlias, ClassVar, Any, TypeVar, Coroutine
from enum import Enum, IntEnum

from yarl import URL

_T = TypeVar("_T")

CORO: TypeAlias = Coroutine[Any, Any, _T]


# https://stackoverflow.com/a/54732120/19419998
class Game(Enum):
    CSGO = 730, 2
    DOTA2 = 570, 2
    H1Z1 = 433850, 2
    RUST = 252490, 2
    TF2 = 440, 2
    PUBG = 578080, 2

    STEAM = 753, 6  # not actually a game :)

    _steam_id_map: ClassVar[dict[int, "Game"]]

    def __new__(cls, *args, **kwargs):
        obj = object.__new__(cls)
        obj._value_ = args[0]
        return obj

    def __init__(self, _, context_id):
        self._context_id_ = context_id
        self._args_tuple_ = (self._value_, self._context_id_)

    @property
    def context_id(self) -> int:
        return self._context_id_

    @property
    def app_id(self) -> int:
        return self._value_

    @classmethod
    def by_steam_id(cls, steam_id: int) -> "Game | None":
        return cls._steam_id_map.get(steam_id)

    def __getitem__(self, index: int) -> int:
        return self._args_tuple_[index]

    def __iter__(self):  # for unpacking
        return iter(self._args_tuple_)


Game._steam_id_map = {g.value: g for g in Game.__members__.values()}

GameType: TypeAlias = Game | tuple[int, int]


class Currency(IntEnum):
    """
    Steam currency enum.

    .. seealso:: https://partner.steamgames.com/doc/store/pricing/currencies
    """

    USD = 1  # UnitedStates Dollar
    GBP = 2  # United Kingdom Pound
    EURO = 3  # European Union Euro
    CHF = 4  # Swiss Francs
    RUB = 5  # Russian Rouble
    PLN = 6  # Polish Złoty
    BRL = 7  # Brazilian Reals
    JPY = 8  # Japanese Yen
    NOK = 9  # Norwegian Krone
    IDR = 10  # Indonesian Rupiah
    MYR = 11  # Malaysian Ringgit
    PHP = 12  # Philippine Peso
    SGD = 13  # Singapore Dollar
    THB = 14  # Thai Baht
    VND = 15  # Vietnamese Dong
    KRW = 16  # South KoreanWon
    TRY = 17  # Turkish Lira
    UAH = 18  # Ukrainian Hryvnia
    MXN = 19  # Mexican Peso
    CAD = 20  # Canadian Dollars
    AUD = 21  # Australian Dollars
    NZD = 22  # New Zealand Dollar
    CNY = 23  # Chinese Renminbi (yuan)
    INR = 24  # Indian Rupee
    CLP = 25  # Chilean Peso
    PEN = 26  # Peruvian Sol
    COP = 27  # Colombian Peso
    ZAR = 28  # South AfricanRand
    HKD = 29  # Hong KongDollar
    TWD = 30  # New TaiwanDollar
    SAR = 31  # Saudi Riyal
    AED = 32  # United ArabEmirates Dirham
    # SEK = 33  # Swedish Krona
    ARS = 34  # Argentine Peso
    ILS = 35  # Israeli NewShekel
    # BYN = 36  # Belarusian Ruble
    KZT = 37  # Kazakhstani Tenge
    KWD = 38  # Kuwaiti Dinar
    QAR = 39  # Qatari Riyal
    CRC = 40  # Costa Rican Colón
    UYU = 41  # Uruguayan Peso
    # BGN = 42  # Bulgarian Lev
    # HRK = 43  # Croatian Kuna
    # CZK = 44  # Czech Koruna
    # DKK = 45  # Danish Krone
    # HUF = 46  # Hungarian Forint
    # RON = 47  # Romanian Leu

    _name_map: ClassVar[dict[str, "Currency"]]

    @classmethod
    def by_name(cls, name: str) -> "Currency":
        return cls._name_map[name]


Currency._name_map = {c.name: c for c in Currency.__members__.values()}


class Language(str, Enum):  # like StrEnum from python 3.11
    """
    Steam languages.

    .. seealso:: https://partner.steamgames.com/doc/store/localization/languages
    """

    ARABIC = "arabic"
    BULGARIAN = "bulgarian"
    SIMPLIFIED_CHINESE = "schinese"
    TRADITIONAL_CHINESE = "tchinese"
    CZECH = "czech"
    DANISH = "danish"
    DUTCH = "dutch"
    ENGLISH = "english"
    FINNISH = "finnish"
    FRENCH = "french"
    GERMAN = "german"
    GREEK = "greek"
    HUNGARIAN = "hungarian"
    ITALIAN = "italian"
    JAPANESE = "japanese"
    KOREAN = "koreana"
    NORWEGIAN = "norwegian"
    POLISH = "polish"
    PORTUGUESE = "portuguese"
    PORTUGUESE_BRAZIL = "brazilian"
    ROMANIAN = "romanian"
    RUSSIAN = "russian"
    SPANISH = "spanish"
    SPANISH_LATIN_AMERICAN = "latam"
    SWEDISH = "swedish"
    THAI = "thai"
    TURKISH = "turkish"
    UKRAINIAN = "ukrainian"
    VIETNAMESE = "vietnamese"

    def __str__(self):
        return self.value


# TODO Countries enum


class TradeOfferStatus(Enum):
    INVALID = 1
    ACTIVE = 2
    ACCEPTED = 3
    COUNTERED = 4
    EXPIRED = 5
    CANCELED = 6
    DECLINED = 7
    INVALID_ITEMS = 8
    CONFIRMATION_NEED = 9
    CANCELED_BY_SECONDARY_FACTOR = 10
    STATE_IN_ESCROW = 11


# https://github.com/DoctorMcKay/node-steamcommunity/blob/master/resources/EConfirmationType.js
class ConfirmationType(Enum):
    UNKNOWN = 1
    TRADE = 2
    LISTING = 3

    @classmethod
    def get(cls, v: int) -> "ConfirmationType":
        try:
            return cls(v)
        except ValueError:
            return cls.UNKNOWN


class MarketListingStatus(Enum):
    NEED_CONFIRMATION = 17
    ACTIVE = 1


class MarketHistoryEventType(Enum):
    LISTING_CREATED = 1
    LISTING_CANCELED = 2
    LISTING_SOLD = 3
    LISTING_PURCHASED = 4


_API_BASE = URL("https://api.steampowered.com")  # nah


class STEAM_URL:
    COMMUNITY = URL("https://steamcommunity.com")
    STORE = URL("https://store.steampowered.com")
    HELP = URL("https://help.steampowered.com")
    STATIC = URL("https://community.akamai.steamstatic.com")
    # specific
    MARKET = COMMUNITY / "market/"
    TRADE = COMMUNITY / "tradeoffer"

    class API:
        BASE = _API_BASE

        # interfaces
        class IEconService:
            _IBase = _API_BASE / "IEconService"
            _v = "v1"
            GetTradeHistory = _IBase / "GetTradeHistory" / _v
            GetTradeHoldDurations = _IBase / "GetTradeHoldDurations" / _v
            GetTradeOffer = _IBase / "GetTradeOffer" / _v
            GetTradeOffers = _IBase / "GetTradeOffers" / _v
            GetTradeOffersSummary = _IBase / "GetTradeOffersSummary" / _v
            GetTradeStatus = _IBase / "GetTradeStatus" / _v
